"""."""
import numpy as _np


def _get_matrix_ss_section(leng):
    return _np.array([
        [1, -leng/2, 0, 0],
        [1, leng/2, 0, 0],
        [0, 0, 1, -leng/2],
        [0, 0, 1, leng/2]])


# NOTE: these matrices translate desired angles and positions of
# the bump at specific longitudinal positions to associated BPM orbit
# distortions. Usually these bumps are specified at source points, dipoles
# or insertion devices, and the BPMs adjacent to these source poiints.
#
# these matrices are calculated using the funcntion 'calc_matrices'
# with corresponding modules in apsuite.orbcorr subpackage.
BUMP_MATRICES = {
    'SA': _get_matrix_ss_section(7.0358),
    'SB': _get_matrix_ss_section(6.1758),
    'SP': _get_matrix_ss_section(6.1758),
    # The BC matrix was calculated using the storage ring model.
    'BC': _np.array([
        [1.11371, -0.61624, 0, 0],
        [1.25316, 1.54265, 0, 0],
        [0, 0, 0.90631, -0.57170],
        [0, 0, 0.69528, 1.80929]]),
    # The B2 matrix was calculated using the storage ring model.
    'C2': _np.array([  # NOTE: first B2 (23.0 mrad) in sector (in subsec C2).
        [1.23679, -0.43310, 0, 0],
        [0.75773, 1.19247, 0, 0],
        [0, 0, 0.90320, -0.50532],
        [0, 0, -0.68264, 6.16011]]),
    # The B1 matrix was calculated using the storage ring model.
    'C1': _np.array([  # NOTE: first B1 (3.2 mrad) in sector (in subsec C1).
        [0.94139, -1.26921, 0, 0],
        [2.58532, 2.83033, 0, 0],
        [0, 0, 0.97724, -1.92058],
        [0, 0, 0.09027, 1.89970]]),
    }


def si_calculate_bump(orbx, orby, subsec, agx=0, agy=0, psx=0, psy=0):
    """Calculate bumps on SA, SB, SP, BC, C2 and C1 subsections.

    Inputs:
        agx, agy, psx, psy - are the desired angles and positions at the
            center of the straight section/center of BC/source point at B2/
            source point at B1. The must have the same unit as orbx and orby.
        subsec - is a 4 character string with the name of the subsection where
            the bump is to be made. Examples: '01SA', '10SA', '13BC', '02SB'
        orbx, orby - the 160-long vector related to the storage ring orbit.
            Ideally this is the reference orbit around which the bump will be
            made.

    Outputs:
        orbx, orby - The orbits with the bump applied to them
    """
    # These are the indices of the BPMs inside one section of the ring:
    bpmidcs = {'BC': 4, 'SA': 0, 'SB': 0, 'SP': 0, 'C2': 3, 'C1': 1}

    # get section and subsection
    sec = int(subsec[:2])
    if not 1 <= sec <= 20:
        raise ValueError('Section must be between 01..20.')
    sec -= 1
    subname = subsec[2:]

    vec = _np.array([psx, agx, psy, agy])
    pos_bpm = _np.dot(BUMP_MATRICES[subname], vec)

    bpm1 = sec * 8 + bpmidcs[subname] - 1
    bpm2 = sec * 8 + bpmidcs[subname]

    orbx, orby = orbx.copy(), orby.copy()
    orbx[bpm1] += pos_bpm[0]
    orbx[bpm2] += pos_bpm[1]
    orby[bpm1] += pos_bpm[2]
    orby[bpm2] += pos_bpm[3]
    return orbx, orby



#!/usr/bin/env python-sirius

"""Unittest module for ll_time_search.py."""

from unittest import TestCase
from siriuspy import util
from siriuspy.timesys import plot_network

mock_flag = True

public_interface = ('PlotNetwork', )


class TestModule(TestCase):
    """Test module interface."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
                                            plot_network, public_interface)
        self.assertTrue(valid)


class TestPlotNetwork(TestCase):
    """Test PlotNetwork class."""

    public_interface = ('plot', )

    def test_public_interface(self):
        """Test class public interface."""
        valid = util.check_public_interface_namespace(
            plot_network.PlotNetwork, TestPlotNetwork.public_interface)
        self.assertTrue(valid)

    def test_plot(self):
        """Test plot."""
        # TODO: implement test!
        pass



"""BLM Search module."""

from ..namesys import Filter as _Filter
from ..namesys import SiriusPVName as _SiriusPVName


class BLMSearch:
    """Beam Loss Monitor Search Class."""

    _blmname2data = {
        'SENSOR1': ('SI-09C3', 'RIGHT', 'Ch7'),
        'SENSOR2': ('SI-09C3', 'LEFT', 'Ch8'),
        'SENSOR3': ('SI-10M1', 'BOTTOM', 'Ch7'),
        'SENSOR4': ('SI-10M1', 'TOP', 'Ch8'),
        'SENSOR5': ('SI-18C3', 'RIGHT', 'Ch7'),
        'SENSOR6': ('SI-18C3', 'LEFT', 'Ch8'),
        'SENSOR7': ('SI-19M1', 'BOTTOM', 'Ch7'),
        'SENSOR8': ('SI-19M1', 'TOP', 'Ch8'),
    }

    @staticmethod
    def get_blmnames(filters=None):
        """Return a sorted and filtered list of all BLM names."""
        blmnames_list = list(BLMSearch._blmname2data.keys())
        blmnames = _Filter.process_filters(blmnames_list, filters=filters)
        return sorted(blmnames)

    @staticmethod
    def conv_blmname_2_secsub(blmname):
        """Return the sector-subsector where BLM is located."""
        sec = BLMSearch._blmname2data[blmname][0]
        return sec

    @staticmethod
    def conv_blmname_2_position(blmname):
        """Return the position of the BLM with sector."""
        pos = BLMSearch._blmname2data[blmname][1]
        return pos

    @staticmethod
    def conv_blmname_2_counter_channel(blmname):
        """Return the counter channel of a BLM."""
        chn = BLMSearch._blmname2data[blmname][2]
        return chn

    @staticmethod
    def conv_blmname_2_devname(blmname):
        """Return device name of a BLM."""
        sec, _, chn = BLMSearch._blmname2data[blmname]
        devname = _SiriusPVName(sec + ':' + 'CO-Counter-' + chn)
        return devname

    @staticmethod
    def conv_devname_2_blmname(devname):
        """Return BLM name from device name."""
        for blmname in BLMSearch._blmname2data:
            sec, _, chn = BLMSearch._blmname2data[blmname]
            devname_ = _SiriusPVName(sec + ':' + 'CO-Counter-' + chn)
            if devname_ == devname:
                return blmname
        raise ValueError(f'Invalid devname {devname}')



"""Devices to be used in power supply procedures."""

import time as _time
import numpy as _np

from ..util import get_bit as _get_bit
from ..namesys import SiriusPVName as _PVName
from ..search import PSSearch as _PSSearch
from ..pwrsupply.csdev import Const as _PSC, ETypes as _PSE, \
    PS_LI_INTLK_THRS as _PS_LI_INTLK, \
    get_ps_interlocks as _get_ps_interlocks

from .device import Device as _Device, DeviceSet as _DeviceSet
from .timing import Trigger as _Trigger


DEFAULT_CAP_BANK_VOLT = {
    'FBP_DCLink': 100,
    'PA-RaPSE01:PS-DCLink-BO': 240,
    'PA-RaPSE02:PS-DCLink-BO': 240,
    'PA-RaPSE03:PS-DCLink-BO': 240,
    'PA-RaPSE04:PS-DCLink-BO': 240,
    'PA-RaPSE06:PS-DCLink-BO': 240,
    'PA-RaPSE07:PS-DCLink-BO': 240,
    'PA-RaPSE08:PS-DCLink-BO': 240,
    'PA-RaPSE09:PS-DCLink-BO': 240,
    'PA-RaPSF01:PS-DCLink-BO': 240,
    'PA-RaPSF02:PS-DCLink-BO': 240,
    'PA-RaPSF03:PS-DCLink-BO': 240,
    'PA-RaPSF04:PS-DCLink-BO': 240,
    'PA-RaPSF06:PS-DCLink-BO': 240,
    'PA-RaPSF07:PS-DCLink-BO': 240,
    'PA-RaPSF08:PS-DCLink-BO': 240,
    'PA-RaPSF09:PS-DCLink-BO': 240,
    'PA-RaPSC01:PS-DCLink-BO': 300,
    'PA-RaPSC02:PS-DCLink-BO': 300,
    }


class _TesterBase(_Device):
    """Tester base."""


class _TesterPSBase(_TesterBase):
    """Tester PS base."""

    def reset(self):
        """Reset."""
        self['Reset-Cmd'] = 1

    def set_opmode(self, state):
        """Set OpMode."""
        self['OpMode-Sel'] = state

    def check_opmode(self, state):
        """Check OpMode."""
        if isinstance(state, (list, tuple)):
            return self['OpMode-Sts'] in state
        return self['OpMode-Sts'] == state

    def set_pwrstate(self, state='on'):
        """Set PwrState."""
        if state == 'on':
            state = _PSC.PwrStateSel.On
        else:
            state = _PSC.PwrStateSel.Off
        self['PwrState-Sel'] = state

    def check_pwrstate(self, state='on'):
        """Check PwrState."""
        if state == 'on':
            is_ok = self['PwrState-Sts'] == _PSC.PwrStateSts.On
            is_ok &= self['OpMode-Sts'] == _PSC.States.SlowRef
        else:
            is_ok = self['PwrState-Sts'] == _PSC.PwrStateSts.Off
            is_ok &= self['OpMode-Sts'] == _PSC.States.Off
        return is_ok

    def set_ctrlloop(self):
        """Set CtrlLoop."""
        self['CtrlLoop-Sel'] = _PSC.OpenLoop.Closed

    def check_ctrlloop(self):
        """Check CtrlLoop."""
        return self['CtrlLoop-Sts'] == _PSC.OpenLoop.Closed


class TesterDCLinkFBP(_TesterPSBase):
    """FBP DCLink tester."""

    PROPERTIES_DEFAULT = (
        'Reset-Cmd', 'IntlkSoft-Mon', 'IntlkHard-Mon',
        'OpMode-Sel', 'OpMode-Sts',
        'PwrState-Sel', 'PwrState-Sts',
        'CtrlLoop-Sel', 'CtrlLoop-Sts',
        'Voltage-SP', 'Voltage-Mon',
        )

    def check_intlk(self):
        """Check interlocks."""
        status = (self['IntlkHard-Mon'] == 0)
        status &= (self['IntlkSoft-Mon'] == 0)
        return status

    def check_init_ok(self):
        """Check OpMode in SlowRef."""
        status = (self['OpMode-Sts'] == _PSC.States.SlowRef)
        return status

    def set_ctrlloop(self):
        """Set CtrlLoop."""
        self['CtrlLoop-Sel'] = _PSC.OpenLoop.Open

    def check_ctrlloop(self):
        """Check CtrlLoop."""
        return self['CtrlLoop-Sts'] == _PSC.OpenLoop.Open

    def set_capvolt(self):
        """Set capacitor bank voltage."""
        key = 'FBP_DCLink' if self.devname not in DEFAULT_CAP_BANK_VOLT \
            else self.devname
        self['Voltage-SP'] = DEFAULT_CAP_BANK_VOLT[key]

    def check_capvolt(self):
        """Do not need to check."""
        return True

    def check_status(self):
        """Check Status."""
        return self.check_intlk()

    def check_comm(self):
        """Check Communication."""
        pvo = self.pv_object('PwrState-Sts')
        return not pvo.status != 0


class TesterDCLink(_TesterPSBase):
    """DCLink tester."""

    PROPERTIES_DEFAULT = (
        'Reset-Cmd',
        'OpMode-Sel', 'OpMode-Sts',
        'PwrState-Sel', 'PwrState-Sts',
        'CtrlLoop-Sel', 'CtrlLoop-Sts',
        'CapacitorBankVoltage-SP', 'CapacitorBankVoltageRef-Mon',
        'CapacitorBankVoltage-Mon')

    def __init__(self, devname, props2init='all'):
        """Init."""
        if props2init == 'all':
            props2init = list(TesterDCLink.PROPERTIES_DEFAULT)
            self._intlk_pvs = _get_ps_interlocks(psname=devname)
            props2init.extend(self._intlk_pvs)

        super().__init__(devname, props2init=props2init)

    def check_intlk(self):
        """Check interlocks."""
        status = True
        for intlk in self._intlk_pvs:
            status &= (self[intlk] == 0)
        return status

    def check_init_ok(self):
        """Check OpMode in SlowRef."""
        return self['OpMode-Sts'] == _PSC.States.SlowRef

    def set_capvolt(self):
        """Set capacitor bank voltage."""
        self['CapacitorBankVoltage-SP'] = DEFAULT_CAP_BANK_VOLT[self.devname]

    def check_capvolt(self):
        """Check capacitor bank voltage."""
        return self._cmp(
            self['CapacitorBankVoltage-Mon'],
            DEFAULT_CAP_BANK_VOLT[self.devname])

    def check_status(self):
        """Check status."""
        status = self.check_intlk()
        if self.check_pwrstate():
            status &= self._cmp(
                self['CapacitorBankVoltage-Mon'],
                self['CapacitorBankVoltageRef-Mon'])
        return status

    def check_comm(self):
        """Check communication."""
        pvo = self.pv_object('PwrState-Sts')
        return not pvo.status != 0

    def _cmp(self, value, target):
        if None in [value, target]:
            return False
        return _np.isclose(value, target, rtol=0.05)


class TesterDCLinkRegatron(_TesterBase):
    """DCLink tester."""

    _OPMODE_STS_OFF = 4  # READY
    _OPMODE_STS_ON = 8  # RUN
    PROPERTIES_DEFAULT = (
        'Reset-Cmd', 'GenIntlk-Mon', 'GenWarn-Mon', 'OpMode-Sts',
        'PwrState-Sel', 'PwrState-Sts',
        'Voltage-SP', 'VoltageRef-Mon', 'Voltage-Mon',
        )

    def reset(self):
        """Reset."""
        self['Reset-Cmd'] = 1

    def check_intlk(self):
        """Check interlocks."""
        status = (self['GenIntlk-Mon'] == 0)
        status &= (self['GenWarn-Mon'] == 0)
        return status

    def set_pwrstate(self, state='on'):
        """Set PwrState."""
        if state == 'on':
            state = _PSC.OffOn.On
        else:
            state = _PSC.OffOn.Off
        self['PwrState-Sel'] = state

    def check_pwrstate(self, state='on'):
        """Check PwrState."""
        if state == 'on':
            is_ok = self['PwrState-Sts'] == _PSC.OffOn.On
            is_ok &= self['OpMode-Sts'] == self._OPMODE_STS_ON
        else:
            is_ok = self['PwrState-Sts'] == _PSC.OffOn.Off
            is_ok &= self['OpMode-Sts'] == self._OPMODE_STS_OFF
        return is_ok

    def check_init_ok(self):
        """Check OpMode Ok."""
        return self['OpMode-Sts'] == self._OPMODE_STS_ON

    def check_capvolt(self):
        """Check voltage."""
        return _np.isclose(
            self['Voltage-Mon'], self['VoltageRef-Mon'], rtol=0.05)

    def check_status(self):
        """Check status."""
        status = self.check_intlk()
        if self.check_pwrstate():
            status &= self.check_capvolt()
        return status

    def check_comm(self):
        """Check commucation."""
        pvo_intlk = self.pv_object('GenIntlk-Mon')
        pvo_volt = self.pv_object('Voltage-Mon')
        prob = pvo_intlk.status != 0 or pvo_volt.status != 0
        return not prob


class TesterPS(_TesterPSBase):
    """PS tester."""

    PROPERTIES_DEFAULT = (
        'Reset-Cmd',
        'OpMode-Sel', 'OpMode-Sts',
        'PwrState-Sel', 'PwrState-Sts',
        'CtrlLoop-Sel', 'CtrlLoop-Sts',
        'Current-SP', 'CurrentRef-Mon', 'Current-Mon')

    def __init__(self, devname, props2init='all'):
        """Init."""
        if props2init == 'all':
            props2init = list(TesterPS.PROPERTIES_DEFAULT)
            self._intlk_pvs = _get_ps_interlocks(psname=devname)
            props2init.extend(self._intlk_pvs)

        super().__init__(devname, props2init=props2init)

        splims = _PSSearch.conv_psname_2_splims(devname)
        self.test_current = splims['TSTV']
        self.test_tol = splims['TSTR']

    def check_intlk(self):
        """Check interlocks."""
        status = True
        for intlk in self._intlk_pvs:
            status &= (self[intlk] == 0)
        return status

    def set_current(self, test=False, value=None):
        """Set current."""
        if value is None:
            if test:
                value = self.test_current
            else:
                value = 0
        self['Current-SP'] = value

    def check_current(self, test=False, value=None):
        """Check current."""
        if value is None:
            if test:
                value = self.test_current
            else:
                value = 0
        return self._cmp(self['Current-Mon'], value)

    def check_status(self):
        """Check Status."""
        status = self.check_intlk()
        if self.check_pwrstate():
            status &= self._cmp(
                self['Current-Mon'], self['CurrentRef-Mon'])
        return status

    def check_comm(self):
        """Check Communication."""
        pvo = self.pv_object('PwrState-Sts')
        return not pvo.status != 0

    def _cmp(self, value, target):
        if None in [value, target]:
            return False
        return abs(value - target) < self.test_tol


class TesterPSFBP(TesterPS):
    """PS FBP Tester."""

    PROPERTIES_DEFAULT = TesterPS.PROPERTIES_DEFAULT + \
        ('SOFBMode-Sel', 'SOFBMode-Sts')

    def set_sofbmode(self, state='on'):
        """Set SOFBMode."""
        if state == 'on':
            state = _PSC.OffOn.On
        else:
            state = _PSC.OffOn.Off
        self['SOFBMode-Sel'] = state

    def check_sofbmode(self, state='on'):
        """Check SOFBMode."""
        if state == 'on':
            state = _PSC.OffOn.On
        else:
            state = _PSC.OffOn.Off
        return self['SOFBMode-Sts'] == state


class TesterPSLinac(_TesterBase):
    """Linac PS tester."""

    PROPERTIES_DEFAULT = (
        'StatusIntlk-Mon', 'IntlkWarn-Mon',
        'PwrState-Sel', 'PwrState-Sts',
        'Current-SP', 'Current-Mon',
        'Connected-Mon',
        )

    def __init__(self, devname, props2init='all'):
        """."""
        super().__init__(devname, props2init=props2init)

        self.intlkwarn_bit = _PSE.LINAC_INTLCK_WARN.index('LoadI Over Thrs')

        splims = _PSSearch.conv_psname_2_splims(devname)
        self.test_current = splims['TSTV']
        self.test_tol = splims['TSTR']

    def check_intlk(self):
        """Check interlocks."""
        intlkval = self['StatusIntlk-Mon']
        if self.devname.dev == 'Spect':
            intlkwarn = self['IntlkWarn-Mon']
            if _get_bit(intlkwarn, self.intlkwarn_bit):
                intlkval -= 2**self.intlkwarn_bit
        return intlkval < _PS_LI_INTLK

    def set_pwrstate(self, state='on'):
        """Set PwrState."""
        if state == 'on':
            state = _PSC.PwrStateSel.On
        else:
            state = _PSC.PwrStateSel.Off
        self['PwrState-Sel'] = state

    def check_pwrstate(self, state='on'):
        """Check PwrState."""
        if state == 'on':
            state = _PSC.PwrStateSel.On
        else:
            state = _PSC.PwrStateSel.Off
        return self['PwrState-Sts'] == state

    def set_current(self, test=False, value=None):
        """Set current."""
        if value is None:
            if test:
                value = self.test_current
            else:
                value = 0
        self['Current-SP'] = value

    def check_current(self, test=False, value=None):
        """Check current."""
        if value is None:
            value = self.test_current if test else 0
        return self._cmp(self['Current-Mon'], value)

    def check_status(self):
        """Check Status."""
        status = self.check_intlk()
        if self.check_pwrstate():
            status &= self._cmp(
                self['Current-Mon'], self['Current-SP'])
        return status

    def check_comm(self):
        """Check Comminication."""
        pvo = self.pv_object('Connected-Mon')
        prob = pvo.value != 0 or pvo.status != 0
        return not prob

    def _cmp(self, value, target):
        if None in [value, target]:
            return False
        tol = self.test_tol if target else 10*self.test_tol
        return abs(value - target) < tol


class TesterPSFOFB(_TesterBase):
    """FOFB PS tester."""

    PROPERTIES_DEFAULT = (
        'AlarmsAmpLtc-Mon', 'AlarmsAmpLtcRst-Cmd',
        'PwrState-Sel', 'PwrState-Sts',
        'Current-SP', 'CurrentRef-Mon', 'Current-Mon',
        'OpMode-Sel', 'OpMode-Sts',
        )

    def __init__(self, devname, props2init='all'):
        """Init."""
        super().__init__(devname, props2init=props2init)

        splims = _PSSearch.conv_psname_2_splims(devname)
        self.test_current = splims['TSTV']
        self.test_tol = splims['TSTR']

    def reset(self):
        """Reset."""
        self['AlarmsAmpLtcRst-Cmd'] = 1

    def set_opmode(self, state):
        """Set opmode to 'state'."""
        self['OpMode-Sel'] = state

    def check_opmode(self, state):
        """Check whether power supply is in opmode 'state'."""
        if isinstance(state, (list, tuple)):
            return self['OpMode-Sts'] in state
        return self['OpMode-Sts'] == state

    def check_intlk(self):
        """Check interlocks."""
        return self['AlarmsAmpLtc-Mon'] == 0

    def set_pwrstate(self, state='on'):
        """Set PwrState."""
        value = _PSC.OffOn.On if state == 'on' else _PSC.OffOn.Off
        self['PwrState-Sel'] = value

    def check_pwrstate(self, state='on'):
        """Check PwrState."""
        value = _PSC.OffOn.On if state == 'on' else _PSC.OffOn.Off
        return self['PwrState-Sts'] == value

    def set_current(self, test=False, value=None):
        """Set current."""
        if value is None:
            if test:
                value = self.test_current
            else:
                value = 0
        self['Current-SP'] = value

    def check_current(self, test=False, value=None):
        """Check current."""
        if value is None:
            if test:
                value = self.test_current
            else:
                value = 0
        return self._cmp(self['Current-Mon'], value)

    def check_status(self):
        """Check Status."""
        status = self.check_intlk()
        if self.check_pwrstate():
            status &= self._cmp(
                self['Current-Mon'], self['CurrentRef-Mon'])
        return status

    def check_comm(self):
        """Check Communication."""
        return self.pv_object('PwrState-Sts').connected

    def _cmp(self, value, target):
        if None in [value, target]:
            return False
        return abs(value - target) < self.test_tol


class _TesterPUBase(_TesterBase):
    """Tester PU base."""

    PROPERTIES_DEFAULT = (
        'Reset-Cmd',
        'Intlk1-Mon', 'Intlk2-Mon', 'Intlk3-Mon', 'Intlk4-Mon',
        'Intlk5-Mon', 'Intlk6-Mon', 'Intlk7-Mon',
        'PwrState-Sel', 'PwrState-Sts',
        'Pulse-Sel', 'Pulse-Sts',
        'Voltage-SP', 'Voltage-RB', 'Voltage-Mon',
        )

    def __init__(self, devname, props2init='all'):
        """Init."""
        super().__init__(devname, props2init=props2init)

        splims = _PSSearch.conv_psname_2_splims(devname)
        self.test_voltage = splims['TSTV']
        self.test_tol = splims['TSTR']

    def reset(self):
        """Reset."""
        # if there is no interlock, do not send reset to avoid strange IOC
        # behavior (Pulse-Sts is enabled when there are no interlock)
        if self.check_intlk():
            return
        self['Reset-Cmd'] = 1

    def check_intlk(self):
        """Check interlocks."""
        raise NotImplementedError

    def set_pulse(self, state='on'):
        """Set Pulse."""
        # if pulsed magnet was in interlock state and was reset,
        # we need to send a Pulse-Sel = Off before continue
        if self['Pulse-Sel'] == _PSC.OffOn.On \
                and self.check_pulse('off'):
            self['Pulse-Sel'] = _PSC.OffOn.Off
            _time.sleep(0.5)

        if state == 'on':
            state = _PSC.OffOn.On
        else:
            state = _PSC.OffOn.Off
        self['Pulse-Sel'] = state

    def check_pulse(self, state='on'):
        """Check Pulse."""
        if state == 'on':
            state = _PSC.OffOn.On
        else:
            state = _PSC.OffOn.Off
        return self['Pulse-Sts'] == state

    def set_pwrstate(self, state='on'):
        """Set PwrState."""
        # if pulsed magnet was in interlock state and was reset,
        # we need to send a PwrState-Sel = Off before continue
        if self['PwrState-Sel'] == _PSC.OffOn.On \
                and self.check_pwrstate('off'):
            self['PwrState-Sel'] = _PSC.OffOn.Off
            _time.sleep(0.5)

        if state == 'on':
            state = _PSC.OffOn.On
        else:
            state = _PSC.OffOn.Off
        self['PwrState-Sel'] = state

    def check_pwrstate(self, state='on'):
        """Check PwrState."""
        if state == 'on':
            state = _PSC.OffOn.On
        else:
            state = _PSC.OffOn.Off
        return self['PwrState-Sts'] == state

    def set_voltage(self, test=False):
        """Set voltage."""
        value = self.test_voltage if test else 0
        self['Voltage-SP'] = value

    def check_voltage(self, test=False):
        """Check voltage."""
        value = self.test_voltage if test else 0
        return self._cmp(self['Voltage-Mon'], value)

    def check_status(self):
        """Check status."""
        status = self.check_intlk()
        if self.check_pwrstate():
            status &= self._cmp(
                self['Voltage-Mon'], self['Voltage-RB'])
        return status

    def _cmp(self, value, target):
        if None in [value, target]:
            return False
        tol = self.test_tol if target else self.test_tol*2
        return abs(value - target) < tol


class TesterPUKckr(_TesterPUBase):
    """Kicker tester."""

    PROPERTIES_DEFAULT = _TesterPUBase.PROPERTIES_DEFAULT + ('Intlk8-Mon', )

    def check_intlk(self):
        """Check interlocks."""
        status = (self['Intlk1-Mon'] == 1)
        status &= (self['Intlk2-Mon'] == 1)
        status &= (self['Intlk3-Mon'] == 1)
        status &= (self['Intlk4-Mon'] == 1)
        status &= (self['Intlk5-Mon'] == 1)
        status &= (self['Intlk6-Mon'] == 1)
        status &= (self['Intlk7-Mon'] == 1)
        status &= (self['Intlk8-Mon'] == 1)
        return status


class TesterPUSept(_TesterPUBase):
    """Septum tester."""

    def check_intlk(self):
        """Check interlocks."""
        status = (self['Intlk1-Mon'] == 1)
        status &= (self['Intlk2-Mon'] == 1)
        status &= (self['Intlk3-Mon'] == 1)
        status &= (self['Intlk4-Mon'] == 1)
        status &= (self['Intlk5-Mon'] == 1)
        status &= (self['Intlk6-Mon'] == 1)
        status &= (self['Intlk7-Mon'] == 1)
        return status


class PSTesterFactory:
    """Factory class for PS and PU testers."""

    @staticmethod
    def create(devname):
        """Return tester for PS/PU devname."""
        devname = _PVName(devname)
        if devname.sec == 'LI':
            tester = TesterPSLinac(devname)
        elif _PSSearch.conv_psname_2_psmodel(devname) == 'FOFB_PS':
            tester = TesterPSFOFB(devname)
        elif _PSSearch.conv_psname_2_psmodel(devname) == 'FBP_DCLink':
            tester = TesterDCLinkFBP(devname)
        elif 'bo-dclink' in _PSSearch.conv_psname_2_pstype(devname):
            tester = TesterDCLink(devname)
        elif _PSSearch.conv_psname_2_psmodel(devname) == 'REGATRON_DCLink':
            tester = TesterDCLinkRegatron(devname)
        elif _PSSearch.conv_psname_2_psmodel(devname) == 'FBP':
            tester = TesterPSFBP(devname)
        elif devname.dis == 'PS':
            tester = TesterPS(devname)
        elif devname.dis == 'PU' and 'Kckr' in devname.dev:
            tester = TesterPUKckr(devname)
        elif devname.dis == 'PU' and 'Sept' in devname.dev:
            tester = TesterPUSept(devname)
        else:
            raise NotImplementedError(
                f'There is no Tester defined for {devname}.')
        return tester


class Triggers(_DeviceSet):
    """Triggers."""

    def __init__(self, devnames):
        """Init triggers."""
        devices, devrefs = list(), dict()
        for devname in devnames:
            dev = _Trigger(devname, props2init=('State-Sel', 'State-Sts'))
            devrefs[devname] = dev
            devices.append(dev)
        self._devrefs = devrefs
        super().__init__(devices)

    @property
    def triggers(self):
        """Trigger names."""
        return self._devrefs



"""FamBPM deviceSet."""

import logging as _log
import sys
import time as _time
from copy import deepcopy as _dcopy

import numpy as _np

from ..namesys import SiriusPVName as _PVName
from ..search import BPMSearch as _BPMSearch
from .bpm import BPM
from .device import DeviceSet as _DeviceSet

# from threading import Event as _Flag


class FamBPMs(_DeviceSet):
    """."""

    TIMEOUT = 10
    RFFEATT_MAX = 30
    PROPERTIES_ACQ = BPM.PROPERTIES_ACQ
    PROPERTIES_DEFAULT = BPM.PROPERTIES_DEFAULT
    ALL_MTURN_SIGNALS2ACQ = ('A', 'B', 'C', 'D', 'X', 'Y', 'Q', 'S')

    class DEVICES:
        """."""

        SI = 'SI-Fam:DI-BPM'
        BO = 'BO-Fam:DI-BPM'
        ALL = (BO, SI)

    def __init__(
            self, devname=None, bpmnames=None, ispost_mortem=False,
            props2init='all', mturn_signals2acq=('X', 'Y')):
        """Family of BPMs.

        Args:
            devname (str, optional): Device name. If not provided, defaults to
                DEVICES.SI. Determine the list of BPM names.
            bpmnames ((list, tuple), optional): BPM names list. If provided,
                it takes priority over 'devname' parameter. Defaults to None.
            ispost_mortem (bool, optional):  Whether to control PM acquisition
                core. Defaults to False.
            props2init (str|list|tuple, optional): list of properties to
                initialize in object creation. If a string is passed, it can
                be either "all" or "acq". Defaults to "all".
            mturn_signals2acq (str|list|tuple, optional): Which signals to
                acquire. Signals can be defined by:
                    X: x position;
                    Y: y position;
                    S: antennas sum signal;
                    Q: skew position;
                    A: amplitude of antenna A;
                    B: amplitude of antenna B;
                    C: amplitude of antenna C;
                    D: amplitude of antenna D.
                Defaults to "XY".

        """
        if devname is None:
            devname = self.DEVICES.SI
        if devname not in self.DEVICES.ALL:
            raise ValueError('Wrong value for devname')

        devname = _PVName(devname)
        bpm_names = bpmnames or _BPMSearch.get_names(
            filters={'sec': devname.sec, 'dev': devname.dev})
        self._ispost_mortem = ispost_mortem

        self._mturn_signals2acq = ''
        self.bpms = [BPM(
            dev, auto_monitor_mon=False, ispost_mortem=ispost_mortem,
            props2init=props2init) for dev in bpm_names]

        # use property here to ensure auto_monitor
        self.mturn_signals2acq = list(mturn_signals2acq)

        super().__init__(self.bpms[:], devname=devname)
        self._bpm_names = bpm_names
        self._csbpm = self.bpms[0].csdata
        self._initial_timestamps = None
        self._initial_signals = None

        self._mturn_flags = dict()
        # NOTE: ACQCount-Mon need to be fixed on BPM's IOC
        # for bpm in devs:
        #     pvo = bpm.pv_object('ACQCount-Mon')
        #     pvo.auto_monitor = True
        #     self._mturn_flags[pvo.pvname] = _Flag()
        #     pvo.add_callback(self._set_mturn_flag)

    @property
    def bpm_names(self):
        """Return BPM names."""
        return self._bpm_names

    @property
    def csbpm(self):
        """Return control system BPM constants class."""
        return self._csbpm

    @property
    def mturn_signals2acq(self):
        """Return which signals will be acquired by get_mturn_signals."""
        return _dcopy(self._mturn_signals2acq)

    @mturn_signals2acq.setter
    def mturn_signals2acq(self, sigs):
        sigs = [s.upper() for s in sigs]
        diff = set(sigs) - set(self.ALL_MTURN_SIGNALS2ACQ)
        if diff:
            raise ValueError('The following signals do not exist: '+str(diff))
        self._configure_automonitor_acquisition_pvs(state=False)
        self._mturn_signals2acq = sigs
        self._configure_automonitor_acquisition_pvs(state=True)

    @staticmethod
    def calc_positions_from_amplitudes(
        amps, gainx=None, gainy=None, offx=0.0, offy=0.0
    ):
        """Calculate transverse positions from antennas amplitudes.

        Args:
            amps (list|numpy.ndarray, (4, ...)): A list or N-D array
                containing the amplitudes of the four antennas in the order
                A, B, C, D.
            gainx (float|numpy.ndarray, optional): Horizontal gain in [um]. If
                None, then the nominal value of 12/sqrt(2) of the storage ring
                will be used. Defaults to None. An array can be used to
                specify different values for each BPM. In this case the gain
                array must match the shape of the antennas arrays or satisfy
                the broadcast rules in relation to them.
            gainy (float|numpy.ndarray, optional): Vertical gain in [um]. If
                None, then the nominal value of 12/sqrt(2) of the storage ring
                will be used. Defaults to None. An array can be used to
                specify different values for each BPM. In this case the gain
                array must match the shape of the antennas arrays or satisfy
                the broadcast rules in relation to them.
            offx (float|numpy.ndarray, optional): Horizontal offset in [um].
                Defaults to 0.0. An array can be used to specify different
                values for each BPM. In this case the gain array must match
                the shape of the antennas arrays or satisfy the broadcast
                rules in relation to them.
            offy (float|numpy.ndarray, optional): Vertical offset in [um].
                Defaults to 0.0. An array can be used to specify different
                values for each BPM. In this case the gain array must match
                the shape of the antennas arrays or satisfy the broadcast
                rules in relation to them.

        Returns:
            posx (numpy.ndarray, (...)): Horizontal position in [um]. Same
                shape of antennas array.
            posy (numpy.ndarray, (...)): Vertical position in [um]. Same
                shape of antennas array.

        """
        a, b, c, d = amps

        radius = 12e3  # [um]
        gainx = radius / _np.sqrt(2) if gainx is None else gainx
        gainy = radius / _np.sqrt(2) if gainy is None else gainy
        gainx = _np.atleast_1d(gainx)
        gainy = _np.atleast_1d(gainy)
        offx = _np.atleast_1d(offx)
        offy = _np.atleast_1d(offy)

        # Calculate difference over sum for each pair
        a_c = (a - c) / (a + c)
        b_d = (b - d) / (b + d)
        # Get the positions:
        posx = (a_c - b_d) / 2
        posy = (a_c + b_d) / 2
        # Apply position gains:
        posx *= gainx
        posy *= gainy
        # Subtract offsets:
        posx -= offx
        posy -= offy
        return posx, posy

    def set_attenuation(self, value=RFFEATT_MAX, timeout=TIMEOUT):
        """."""
        for bpm in self:
            bpm.rffe_att = value

        mstr = ''
        okall = True
        t0 = _time.time()
        for bpm in self:
            tout = timeout - (_time.time() - t0)
            if not bpm._wait('RFFEAtt-RB', value, timeout=tout):
                okall = False
                mstr += (
                    f'\n{bpm.devname:<20s}: ' +
                    f'rb {bpm.rffe_att:.0f} != sp {value:.0f}')

        stg = ', except:' if mstr else '.'
        _log.info('RFFE attenuation set confirmed in all BPMs%s', stg)
        if mstr:
            _log.info(mstr)
        return okall

    def get_slow_orbit(self):
        """Get slow orbit vectors.

        Returns:
            orbx (numpy.ndarray, 160): Horizontal Orbit.
            orby (numpy.ndarray, 160): Vertical Orbit.

        """
        orbx, orby = [], []
        for bpm in self.bpms:
            orbx.append(bpm.posx)
            orby.append(bpm.posy)
        orbx = _np.array(orbx)
        orby = _np.array(orby)
        return orbx, orby

    def get_mturn_signals(self):
        """Get Multiturn signals matrices.

        Returns:
            list: Each component of the list is an numpy.ndarray with shape
                (N, 160), containing the values for the signals acquired.

        """
        sigs = [[] for _ in self._mturn_signals2acq]

        mini = int(sys.maxsize)  # a very large integer
        for bpm in self.bpms:
            for i, sn in enumerate(self._mturn_signals2acq):
                sn = 'sum' if sn == 'S' else sn.lower()
                name = 'mt_' + ('ampl' if sn in 'abcd' else 'pos') + sn
                sigs[i].append(getattr(bpm, name))
            mini = min(mini, _np.min([s[-1].size for s in sigs]))

        for i, sig in enumerate(sigs):
            for j, s in enumerate(sig):
                sig[j] = s[:mini]
            sigs[i] = _np.array(sig).T
        return sigs

    def get_mturn_timestamps(self):
        """Get Multiturn data timestamps.

        Returns:
            tsmps (numpy.ndarray, (N, 160)): The i-th column has the timestamp
                of the i-th bpm for the N aquired signals.

        """
        tsmps = _np.zeros(
            (len(self._mturn_signals2acq), len(self.bpms)), dtype=float)
        for i, s in enumerate(self._mturn_signals2acq):
            for j, bpm in enumerate(self.bpms):
                s = 'SUM' if s == 'S' else s
                pvo = bpm.pv_object(f'GEN_{s}ArrayData')
                tv = pvo.get_timevars(timeout=self.TIMEOUT)
                tsmps[i, j] = pvo.timestamp if tv is None else tv['timestamp']
        return tsmps

    def get_sampling_frequency(self, rf_freq: float, acq_rate='') -> float:
        """Return the sampling frequency of the acquisition.

        Args:
            rf_freq (float): RF frequency.
            acq_rate (str, optional): acquisition rate. Defaults to ''.
            If empty string, it gets the configured acq. rate on BPMs

        Returns:
            float: acquisition frequency.

        """
        fs_bpms = {
            dev.get_sampling_frequency(rf_freq, acq_rate)
            for dev in self.bpms}
        if len(fs_bpms) == 1:
            return fs_bpms.pop()
        else:
            _log.warning('BPMs are not configured with the same ACQChannel.')
            return None

    def get_switching_frequency(self, rf_freq: float) -> float:
        """Return the switching frequency.

        Args:
            rf_freq (float): RF frequency.

        Returns:
            float: switching frequency.

        """
        fsw_bpms = {
            dev.get_switching_frequency(rf_freq) for dev in self.bpms}
        if len(fsw_bpms) == 1:
            return fsw_bpms.pop()
        else:
            _log.warning('BPMs are not configured with the same SwMode.')
            return None

    def config_mturn_acquisition(
            self, nr_points_after: int, nr_points_before=0,
            acq_rate='FAcq', repeat=True, external=True) -> int:
        """Configure acquisition for BPMs.

        Args:
            nr_points_after (int): number of points after trigger.
            nr_points_before (int): number of points after trigger.
                Defaults to 0.
            acq_rate (int|str, optional): Acquisition rate name ('TbT',
                'TbTPha', 'FOFB', 'FOFBPha', 'FAcq', 'ADC', 'ADCSwp') or value.
                Defaults to 'FAcq'.
            repeat (bool, optional): Whether or not acquisition should be
                repetitive. Defaults to True.
            external (bool, optional): Whether or not external trigger should
                be used. Defaults to True.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                <0: Index of the first BPM which did not stop last acq. plus 1.
                >0: Index of the first BPM which is not ready for acq. plus 1.

        """
        dic = {
            'facq': self._csbpm.AcqChan.FAcq,
            'fofbpha': self._csbpm.AcqChan.FOFBPha,
            'fofb': self._csbpm.AcqChan.FOFB,
            'tbtpha': self._csbpm.AcqChan.TbTPha,
            'tbt': self._csbpm.AcqChan.TbT,
            'adcswp': self._csbpm.AcqChan.ADCSwp,
            'adc': self._csbpm.AcqChan.ADC,
        }
        if isinstance(acq_rate, str):
            acq_rate = dic.get(acq_rate.lower())
        if acq_rate not in self._csbpm.AcqChan:
            raise ValueError(
                str(acq_rate) + ' is not a valid acquisition rate.')

        rep = self._csbpm.AcqRepeat.Normal
        if repeat:
            rep = self._csbpm.AcqRepeat.Repetitive

        trig = self._csbpm.AcqTrigTyp.Now
        if external:
            trig = self._csbpm.AcqTrigTyp.External

        ret = self.cmd_abort_mturn_acquisition()
        if ret > 0:
            return -ret

        for bpm in self.bpms:
            bpm.acq_repeat = rep
            bpm.acq_channel = acq_rate
            bpm.acq_trigger = trig
            bpm.acq_nrsamples_pre = nr_points_before
            bpm.acq_nrsamples_post = nr_points_after

        return self.cmd_start_mturn_acquisition()

    def cmd_abort_mturn_acquisition(self, wait=True, timeout=10) -> int:
        """Abort BPMs acquistion.

        Args:
            wait (bool, optional): whether or not to wait BPMs get ready.
                Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for bpm in self.bpms:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Abort

        if wait:
            return self.wait_acquisition_finish(timeout=timeout)
        return 0

    def wait_acquisition_finish(self, timeout=10) -> int:
        """Wait for all BPMs to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, bpm in enumerate(self.bpms):
            t0_ = _time.time()
            if not bpm.wait_acq_finish(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def cmd_start_mturn_acquisition(self, wait=True, timeout=10) -> int:
        """Start BPMs acquisition.

        Args:
            wait (bool, optional): whether or not to wait BPMs get ready.
                Defaults to True.
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for bpm in self.bpms:
            bpm.acq_ctrl = self._csbpm.AcqEvents.Start
        if wait:
            return self.wait_acquisition_start(timeout=timeout)
        return 0

    def wait_acquisition_start(self, timeout=10) -> bool:
        """Wait for all BPMs to be ready for acquisition.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, bpm in enumerate(self.bpms):
            t0_ = _time.time()
            if not bpm.wait_acq_start(timeout):
                return i + 1
            timeout -= _time.time() - t0_
        return 0

    def set_switching_mode(self, mode='direct'):
        """Set switching mode of BPMS.

        Args:
            mode ((str, int), optional): Desired mode, must be in
                {'direct', 'switching', 1, 3}. Defaults to 'direct'.

        Raises:
            ValueError: When mode is not in {'direct', 'switching', 1, 3}.

        """
        if mode not in ('direct', 'switching', 1, 3):
            raise ValueError('Value must be in ("direct", "switching", 1, 3).')

        for bpm in self.bpms:
            bpm.switching_mode = mode

    def update_mturn_initial_signals(self):
        """Call this method before acquisition to get orbit for comparison."""
        sig = _np.array(self.get_mturn_signals())
        self._initial_signals = sig if sig.shape[1] != 0 else None

    def update_mturn_initial_timestamps(self):
        """Call this method before acquisition to get orbit for comparison."""
        self._initial_timestamps = self.get_mturn_timestamps()

    def reset_mturn_flags(self):
        """Reset Multiturn flags to wait for a new orbit update."""
        for flag in self._mturn_flags.values():
            flag.clear()

    def reset_mturn_initial_state(self):
        """Set initial state to wait for orbit acquisition to start."""
        self.reset_mturn_flags()
        self.update_mturn_initial_timestamps()
        self.update_mturn_initial_signals()

    def wait_update_mturn_flags(self, timeout=10):
        """Wait for all acquisition flags to be updated.

        Args:
            timeout (int, optional): Time to wait. Defaults to 10.

        Returns:
            int: code describing what happened:
                =0: BPMs are ready.
                >0: Index of the first BPM which did not update plus 1.

        """
        for i, flag in enumerate(self._mturn_flags.values()):
            t00 = _time.time()
            if not flag.wait(timeout=timeout):
                return i + 1
            timeout -= _time.time() - t00
            timeout = max(timeout, 0)
        return 0

    def wait_update_mturn_timestamps(self, timeout=10) -> int:
        """Call this method after acquisition to check for timestamps update.

        For this method to work it is necessary to call
            update_mturn_initial_timestamps
        before the acquisition starts, so that a reference for comparison is
        created.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.

        Returns:
            int|float: code describing what happened:
                -2: size of timestamps changed in relation to initial timestamp
                -1: initial timestamps were not defined;
                =0: data updated.
                >0: timeout waiting BPMs. The returned value is a float of
                    form X.Y, where:
                        X: The integer part indicates the index of the first
                            BPM which did not update plus 1.
                        Y: The fractional part indicates which signal, from
                            1 to N, that didn't update.

        """
        if self._initial_timestamps is None:
            return -1

        tsmp0 = self._initial_timestamps
        while timeout > 0:
            t00 = _time.time()
            tsmp = self.get_mturn_timestamps()
            if tsmp.shape != tsmp0.shape:
                return -2
            errors = _np.equal(tsmp, tsmp0)
            if not _np.any(errors):
                return 0
            _time.sleep(0.1)
            timeout -= _time.time() - t00

        bpm_idx = int(_np.nonzero(_np.any(errors, axis=0))[0][0])
        code = bpm_idx + 1
        code += (int(_np.nonzero(errors[:, bpm_idx])[0][0]) + 1) / 10
        return code

    def wait_update_mturn_signals(self, timeout=10) -> int:
        """Call this method after acquisition to check for data update.

        For this method to work it is necessary to call
            update_mturn_initial_signals
        before the acquisition starts, so that a reference for comparison is
        created.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.

        Returns:
            int|float: code describing what happened:
                -2: size of signals is 0;
                -1: initial signals were not defined;
                =0: signals updated.
                >0: timeout waiting BPMs. The returned value is a float of
                    form X.Y, where:
                        X: The integer part indicates the index of the first
                            BPM which did not update plus 1.
                        Y: The fractional part indicates which signal, from
                            1 to N, that didn't update.

        """
        if self._initial_signals is None:
            return -1

        sig0 = self._initial_signals
        while timeout > 0:
            t00 = _time.time()
            sig = _np.array(self.get_mturn_signals())
            siz = min(sig.shape[1], sig0.shape[1])
            if siz == 0:
                return -2
            errors = _np.all(_np.equal(sig[:, :siz], sig0[:, :siz]), axis=1)
            if not _np.any(errors):
                return 0
            _log.debug('Signals did not update yet. Trying again.')
            _time.sleep(0.1)
            timeout -= _time.time() - t00

        bpm_idx = int(_np.nonzero(_np.any(errors, axis=0))[0][0])
        code = bpm_idx + 1
        code += (int(_np.nonzero(errors[:, bpm_idx])[0][0]) + 1) / 10
        return code

    def wait_update_mturn(self, timeout=10) -> int:
        """Combine all methods to wait update data.

        Args:
            timeout (int, optional): Waiting timeout. Defaults to 10.

        Returns:
            int|float: code describing what happened:
                -2: size of timestamps changed in relation to initial timestamp
                -1: initial timestamps were not defined;
                =0: data updated.
                >0: timeout waiting BPMs. The returned value is a float of
                    form X.Y, where:
                        X: The integer part indicates the index of the first
                            BPM which did not update plus 1.
                        Y: The fractional part indicates which signal, from
                            1 to N, that didn't update.

        """
        t00 = _time.time()
        ret = self.wait_update_mturn_flags(timeout)
        if ret > 0:
            return ret
        t01 = _time.time()
        dtime = t01 - t00
        timeout -= dtime
        _log.debug("Flags updated (ETA: %.3fs).", dtime)

        ret = self.wait_update_mturn_timestamps(timeout)
        if ret != 0:
            return ret
        t02 = _time.time()
        dtime = t02 - t01
        timeout -= dtime
        _log.debug("Timestamps updated (ETA: %.3fs).", dtime)

        ret = self.wait_update_mturn_signals(timeout)
        if ret != 0:
            return ret
        dtime = _time.time() - t02
        _log.debug("Data updated (ETA: %.3fs).", dtime)
        return ret

    # ---------------------- Auxiliary methods ------------------------------
    def _configure_automonitor_acquisition_pvs(self, state):
        for bpm in self.bpms:
            for sig in self._mturn_signals2acq:
                sig = 'SUM' if sig.upper() == 'S' else sig.upper()
                bpm.pv_object(f'GEN_{sig}ArrayData').auto_monitor = state

    def _set_mturn_flag(self, pvname, **kwargs):
        _ = kwargs
        self._mturn_flags[pvname].set()



"""Optics correction utilities module."""
import os as _os


class HandleConfigNameFile:
    """Class to handle config_name file."""

    def __init__(self, acc, opticsparam):
        """Init."""
        if not (acc in ['BO', 'SI']):
            raise ValueError("'acc' must be 'BO' or 'SI' instance.")
        if not (opticsparam in ['tune', 'chrom']):
            raise ValueError(
                "'opticsparam' must be 'tune' or 'chrom' instance.")
        self.acc = acc.lower()
        self.opticsparam = opticsparam
        self.fpath = '/home/sirius/iocs-log/' + self.acc + '-ap-' + \
            opticsparam+'corr/'
        self.fname = self.fpath + self.acc + '-' + opticsparam + 'corr.txt'

    def get_config_name(self):
        """Get the configuration name in file."""
        if _os.path.isfile(self.fname):
            with open(self.fname, 'r') as fil:
                config_name = fil.read().strip('\n')
        else:
            _os.makedirs(self.fpath, exist_ok=True)
            config_name = get_default_config_name(self.acc, self.opticsparam)
            with open(self.fname, 'w+') as fil:
                fil.write(config_name)
        return config_name

    def set_config_name(self, config_name):
        """Set the configuration name in file."""
        if not isinstance(config_name, str):
            raise TypeError("'config_name' must be a string.")
        with open(self.fname, 'w+') as fil:
            fil.write(config_name)


def get_default_config_name(acc, opticsparam):
    """Get default config name."""
    if opticsparam == 'tune':
        if acc.upper() == 'BO':
            return 'BO.V05.04.M0'
        else:
            return 'SI.V24.04_S05.01'
    elif opticsparam == 'chrom':
        if acc.upper() == 'BO':
            return 'BO.V05.04.M0'
        else:
            return 'SI.V24.04_S05.01'



"""Define PVs, constants and properties of LIEnergyMeas IOC."""
import numpy as _np

from .. import csdev as _csdev


# --- Enumeration Types ---
class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    MEASURESTATE = ('Stopped', 'Measuring')


_et = ETypes  # syntactic sugar


# --- Const class ---
class Const(_csdev.Const):
    """Const class defining constants."""

    DEFAULT_DISP = 1087  # in mm
    DEFAULT_B_ANG = _np.pi/4
    DEFAULT_SPECT = 'LI-01:PS-Spect'
    DEFAULT_PROFILE = 'LA-BI:PRF4'

    MeasureState = _csdev.Const.register('MeasureState', _et.MEASURESTATE)

    @classmethod
    def get_database(cls, prefix=''):
        """Return IOC database."""
        dbase = {
            'Dispersion-SP': {
                'type': 'float', 'prec': 4, 'unit': 'mm',
                'value': cls.DEFAULT_DISP},
            'Dispersion-RB': {
                'type': 'float', 'prec': 4, 'unit': 'mm',
                'value': cls.DEFAULT_DISP},
            # 'Angle-SP': {
            #     'type': 'float', 'prec': 4, 'unit': 'deg',
            #     'value': cls.DEFAULT_B_ANG},
            # 'Angle-RB': {
            #     'type': 'float', 'prec': 4, 'unit': 'deg',
            #     'value': cls.DEFAULT_B_ANG},
            # 'Spectrometer-SP': {
            #     'type': 'string', 'value': cls.DEFAULT_SPECT},
            # 'Spectrometer-RB': {
            #     'type': 'string', 'value': cls.DEFAULT_SPECT},
            'IntDipole-Mon': {
                'type': 'float', 'prec': 4, 'unit': 'T.m', 'value': 0},
            'Energy-Mon': {
                'type': 'float', 'prec': 2, 'unit': 'MeV', 'value': 0},
            'Spread-Mon': {
                'type': 'float', 'prec': 4, 'unit': '%', 'value': 0},

            'MeasureCtrl-Sel': {
                'type': 'enum', 'value': cls.MeasureState.Stopped,
                'enums': cls.MeasureState._fields},
            'MeasureCtrl-Sts': {
                'type': 'enum', 'value': cls.MeasureState.Stopped,
                'enums': cls.MeasureState._fields},
            }
        for val in dbase.values():
            low = val.get('lolim', None)
            hig = val.get('hilim', None)
            if low is not None:
                val['lolo'] = low
                val['low'] = low
            if hig is not None:
                val['hihi'] = hig
                val['high'] = hig
        if prefix:
            dbase = {prefix + k: v for k, v in dbase.items()}
        dbase2 = _csdev.Const.get_database(prefix=prefix)
        dbase2[prefix+'ImgFlipY-Sel']['value'] = cls.ImgFlip.Off
        dbase2[prefix+'ImgFlipY-Sts']['value'] = cls.ImgFlip.Off
        dbase2[prefix+'ImgFlipX-Sel']['value'] = cls.ImgFlip.On
        dbase2[prefix+'ImgFlipX-Sts']['value'] = cls.ImgFlip.On
        dbase.update(dbase2)
        return dbase



"""Types subpackage."""

import pkgutil as _pkgutil

CONFIG_TYPES = []

for loader, module_name, is_pkg in _pkgutil.walk_packages(__path__):
    CONFIG_TYPES.append(module_name)
    module = loader.find_module(module_name).load_module(module_name)
    exec('%s = module' % module_name)



"""Main module of AS-AP-ChromCorr IOC."""

import time as _time
from threading import Thread as _Thread
import numpy as _np
from epics import PV as _PV
from epics.ca import ChannelAccessGetFailure as _ChannelAccessGetFailure

from ..envars import VACA_PREFIX as _vaca_prefix
from ..namesys import SiriusPVName as _SiriusPVName
from ..devices import RFGen

from .csdev import Const as _Const
from .base import BaseApp as _BaseApp

_MOM_COMPACT = 1.68e-4


class ChromCorrApp(_BaseApp):
    """Main application for handling chromaticity correction."""

    _optics_param = 'chrom'

    def __init__(self, acc):
        """Class constructor."""
        super().__init__(acc)

        # consts
        self._chrom_sp = 2*[0.0]
        self._deltachrom_sp = 2*[0.0]
        self._chrom_mon = 2*[0.0]
        self._last_param_set = 'abs'

        if self._acc == 'SI':
            self._measuring_chrom = False
            self._meas_chrom_status = _Const.MeasMon.Idle
            self._meas_chrom_lastcomm = _Const.MeasCmd.Reset
            self._meas_chrom_dfreq_rf = 200.0
            self._meas_chrom_wait_tune = 5.0
            self._meas_chrom_nrsteps = 8

            self._meas_config_dsl_sf = 10.000
            self._meas_config_dsl_sd = 10.000

        # Connect to Sextupoles Families
        self._lastcalc_sl = {fam: 0 for fam in self._psfams}
        for fam in self._psfams:
            pvname = _SiriusPVName(self._acc+'-Fam:PS-'+fam+':SL-RB')
            pvname = pvname.substitute(prefix=_vaca_prefix)
            self._psfam_intstr_rb_pvs[fam] = _PV(
                pvname,
                callback=self._callback_estimate_chrom,
                connection_timeout=0.05)

        if self._acc == 'SI':
            # Connect to SI RF
            self._rf_conn = RFGen()

        self.map_pv2write.update({
            'ChromX-SP': self.set_chrom_x,
            'ChromY-SP': self.set_chrom_y,
            'DeltaChromX-SP': self.set_deltachrom_x,
            'DeltaChromY-SP': self.set_deltachrom_y,
            'MeasChromDeltaFreqRF-SP': self.set_meas_chrom_dfreq_rf,
            'MeasChromWaitTune-SP': self.set_meas_chrom_wait_tune,
            'MeasChromNrSteps-SP': self.set_meas_chrom_nrsteps,
            'MeasChrom-Cmd': self.cmd_meas_chrom,
            'MeasConfigDeltaSLFamSF-SP': self.set_meas_config_dsl_sf,
            'MeasConfigDeltaSLFamSD-SP': self.set_meas_config_dsl_sd,
        })

        # auxiliar dict limit names
        self._lims_map = {
            'hilim': 'upper_disp_limit', 'lolim': 'lower_disp_limit',
            'high': 'upper_alarm_limit', 'low': 'lower_alarm_limit',
            'hihi': 'upper_warning_limit', 'lolo': 'lower_warning_limit'}

    def update_corrparams_pvs(self):
        """Set initial correction parameters PVs values."""
        self.run_callbacks('RespMat-Mon', self._nominal_matrix)
        self.run_callbacks('NominalSL-Mon', self._psfam_nom_intstr)
        self.run_callbacks('NominalChrom-Mon', self._nominal_opticsparam)

    def process(self, interval):
        """Sleep."""
        t_ini = _time.time()
        if (self._status & 0x1) == 0:  # Check connection
            for fam in self._psfams:
                rb_pv = self._psfam_intstr_rb_pvs[fam]
                try:
                    data = rb_pv.get_ctrlvars(timeout=0.1)
                    upper_disp_limit = rb_pv.upper_disp_limit
                except _ChannelAccessGetFailure:
                    break
                else:
                    if upper_disp_limit is not None and data is not None:
                        lis = {k: data[v] for k, v in self._lims_map.items()}
                        self.run_callbacks(
                            'SL'+fam+'-Mon', info=lis, field='info')
        dtime = interval - (_time.time() - t_ini)
        _time.sleep(max(dtime, 0))

    # ------ handle pv write methods -------

    def set_chrom_x(self, value):
        """Set ChromX."""
        self._last_param_set = 'absolut'
        self._chrom_sp[0] = value
        self.run_callbacks('ChromX-RB', value)
        self._deltachrom_sp[0] = self._chrom_sp[0]-self._optprm_est[0]
        self._calc_intstrength()
        return True

    def set_chrom_y(self, value):
        """Set ChromY."""
        self._last_param_set = 'absolut'
        self._chrom_sp[1] = value
        self.run_callbacks('ChromY-RB', value)
        self._deltachrom_sp[1] = self._chrom_sp[1]-self._optprm_est[1]
        self._calc_intstrength()
        return True

    def set_deltachrom_x(self, value):
        """Set DeltaChromX."""
        self._last_param_set = 'delta'
        self._deltachrom_sp[0] = value
        self.run_callbacks('DeltaChromX-RB', value)
        self._calc_intstrength()
        return True

    def set_deltachrom_y(self, value):
        """Set DeltaChromY."""
        self._last_param_set = 'delta'
        self._deltachrom_sp[1] = value
        self.run_callbacks('DeltaChromY-RB', value)
        self._calc_intstrength()
        return True

    def set_meas_chrom_dfreq_rf(self, value):
        """Set MeasChromDeltaFreqRF."""
        if value == self._meas_chrom_dfreq_rf:
            return False
        self._meas_chrom_dfreq_rf = value
        self.run_callbacks('MeasChromDeltaFreqRF-RB', value)
        return True

    def set_meas_chrom_wait_tune(self, value):
        """Set MeasChromWaitTune."""
        if value == self._meas_chrom_wait_tune:
            return False
        self._meas_chrom_wait_tune = value
        self.run_callbacks('MeasChromWaitTune-RB', value)
        return True

    def set_meas_chrom_nrsteps(self, value):
        """Set MeasChromNrSteps."""
        if value == self._meas_chrom_nrsteps:
            return False
        self._meas_chrom_nrsteps = value
        self.run_callbacks('MeasChromNrSteps-RB', value)
        return True

    def cmd_meas_chrom(self, value):
        """MeasChrom command."""
        if value == _Const.MeasCmd.Start:
            status = self._start_meas_chrom()
        elif value == _Const.MeasCmd.Stop:
            status = self._stop_meas_chrom()
        elif value == _Const.MeasCmd.Reset:
            status = self._reset_meas_chrom()
        if status:
            self._meas_chrom_lastcomm = value
        return status

    def set_meas_config_dsl_sf(self, value):
        """Set MeasConfigDeltaSLFamSF."""
        if value == self._meas_config_dsl_sf:
            return False
        self._meas_config_dsl_sf = value
        self.run_callbacks('MeasConfigDeltaSLFamSF-RB', value)
        return True

    def set_meas_config_dsl_sd(self, value):
        """Set MeasConfigDeltaSLFamSD."""
        if value == self._meas_config_dsl_sd:
            return False
        self._meas_config_dsl_sd = value
        self.run_callbacks('MeasConfigDeltaSLFamSD-RB', value)
        return True

    # ---------- auxiliar methods ----------

    def _handle_corrparams_2_read(self, params):
        """Edit correction params."""
        nom_matrix = [item for sublist in params['matrix'] for item in sublist]
        nom_sl = params['nominal SLs']
        nom_chrom = params['nominal chrom']
        return nom_matrix, nom_sl, nom_chrom

    def _handle_corrparams_2_save(self):
        matrix = _np.array(self._nominal_matrix)
        matrix = _np.reshape(matrix, [2, len(self._psfams)])

        value = {'matrix': matrix,
                 'nominal SLs': self._psfam_nom_intstr,
                 'nominal chrom': self._nominal_opticsparam}
        return value

    def _calc_intstrength(self):
        if self._last_param_set == 'absolut':
            delta_chromx = self._chrom_sp[0]-self._optprm_est[0]
            delta_chromy = self._chrom_sp[1]-self._optprm_est[1]
        else:
            delta_chromx, delta_chromy = self._deltachrom_sp

        method = 0 \
            if self._corr_method == _Const.CorrMeth.Proportional \
            else 1
        grouping = '2knobs' \
            if self._corr_group == _Const.CorrGroup.TwoKnobs \
            else 'svd'
        lastcalc_deltasl = self._opticscorr.calculate_delta_intstrengths(
            method=method, grouping=grouping,
            delta_opticsparam=[delta_chromx, delta_chromy])

        for fam_idx, fam in enumerate(self._psfams):
            sl_now = self._psfam_intstr_rb_pvs[fam].get()
            if sl_now is None:
                self.run_callbacks(
                    'Log-Mon',
                    'ERR: Received a None value from {}'.format(fam))
                return False
            self._lastcalc_sl[fam] = sl_now + lastcalc_deltasl[fam_idx]
            self.run_callbacks('SL'+fam+'-Mon', self._lastcalc_sl[fam])
        self._estimate_calc_chrom()

        self.run_callbacks('Log-Mon', 'Calculated SL values.')
        return True

    def _apply_corr(self):
        if self._is_status_ok():
            self._apply_intstrength(self._lastcalc_sl)
            self.run_callbacks('Log-Mon', 'Applied correction.')

            if self._sync_corr == _Const.SyncCorr.On:
                self._event_exttrig_cmd.put(0)
                self.run_callbacks('Log-Mon', 'Generated trigger.')
            return True

        self.run_callbacks('Log-Mon', 'ERR:ApplyDelta-Cmd failed.')
        return False

    def _get_optics_param(self):
        """Return optics parameter."""
        sts = self._meas_chrom()
        return sts, self._chrom_mon

    def _get_delta_intstrength(self, fam):
        if 'SL' in fam:
            deltasl = self._meas_config_dsl_sf
        else:
            deltasl = self._meas_config_dsl_sd
        fam_idx = self._psfams.index(fam)
        nelm = self._psfam_nelm[fam_idx]
        return deltasl/nelm

    def _get_rf_freq(self):
        """Get RF Frequnecy."""
        if not self._rf_conn.connected:
            return 0.0
        return self._rf_conn.frequency

    def _start_meas_chrom(self):
        """Start chromaticity measurement."""
        cont = True
        if self._measuring_chrom:
            log_msg = 'ERR: Chrom measurement already in progress!'
            cont = False
        elif not self._tune_x_pv.connected or not self._tune_y_pv.connected:
            log_msg = 'ERR: Cannot measure, tune PVs not connected!'
            cont = False
        elif not self._rf_conn.connected:
            log_msg = 'ERR: Cannot measure, RF PVs not connected!'
            cont = False
        elif not self._is_storedebeam:
            log_msg = 'ERR: Cannot measure, there is no stored beam!'
            cont = False
        if not cont:
            self.run_callbacks('Log-Mon', log_msg)
            return False

        thread = _Thread(target=self._meas_chrom, daemon=True)
        thread.start()
        return True

    def _stop_meas_chrom(self):
        """Stop chromaticity measurement."""
        if not self._measuring_chrom:
            self.run_callbacks(
                'Log-Mon', 'ERR: No chrom measurement occuring!')
            return False
        self.run_callbacks('Log-Mon', 'Aborting chrom measurement!')
        self._measuring_chrom = False
        return True

    def _reset_meas_chrom(self):
        """Reset chromaticity measurement."""
        if self._measuring_chrom:
            self.run_callbacks(
                'Log-Mon', 'ERR: Status not reset, measure in progress!')
            return False
        self.run_callbacks(
            'Log-Mon', 'Reseting chrom measurement status!')
        self._meas_chrom_status = _Const.MeasMon.Idle
        self.run_callbacks('MeasChromStatus-Mon', self._meas_chrom_status)
        return True

    def _meas_chrom(self):
        """Measure chromaticities."""
        self._meas_chrom_status = _Const.MeasMon.Measuring
        self.run_callbacks('MeasChromStatus-Mon', self._meas_chrom_status)

        self._measuring_chrom = True
        self.run_callbacks('Log-Mon', 'Starting chrom measurement!')

        _, data = self._get_tunes()
        tunex0, tuney0 = data

        freq0 = self._get_rf_freq()
        frequencies = _np.linspace(
            freq0-self._meas_chrom_dfreq_rf/2,
            freq0+self._meas_chrom_dfreq_rf/2,
            self._meas_chrom_nrsteps)
        freq_list = list()
        tunex_list, tuney_list = list(), list()

        aborted = False
        for value in frequencies:
            if not self._measuring_chrom:
                log_msg = 'Stoped chrom measurement!'
                aborted = True
            elif not self._is_storedebeam:
                log_msg = 'ERR: Stoping chrom measurement, '\
                          'there is no stored beam!'
                aborted = True
            elif not self._tune_x_pv.connected or \
                    not self._tune_y_pv.connected:
                log_msg = 'ERR: Stoping chrom measurement, '\
                          'tune PVs not connected!'
                aborted = True
            elif not self._rf_conn.connected:
                log_msg = 'ERR: Stoping chrom measurement, '\
                          'RF PVs not connected!'
                aborted = True
            if aborted:
                break

            self._rf_conn.frequency = value
            freq = self._get_rf_freq()
            freq_list.append(freq)
            self.run_callbacks(
                'Log-Mon', 'Delta RF: {} Hz'.format((freq-freq0)))

            _time.sleep(self._meas_chrom_wait_tune)
            sts, data = self._get_tunes()
            if sts:
                tunex, tuney = data
                tunex_list.append(tunex)
                tuney_list.append(tuney)
                self.run_callbacks(
                    'Log-Mon', 'Delta Tune X: {}'.format(
                        (tunex_list[-1]-tunex0)))
                self.run_callbacks(
                    'Log-Mon', 'Delta Tune Y: {}'.format(
                        (tuney_list[-1]-tuney0)))
            else:
                log_msg = 'ERR: Could not measure tune!'
                aborted = True
                break

        self.run_callbacks('Log-Mon', 'Restoring RF frequency...')
        self._rf_conn.frequency = freq0
        self.run_callbacks('Log-Mon', 'RF frequency restored!')

        if aborted:
            self._meas_chrom_status = _Const.MeasMon.Aborted
        else:
            den = -(_np.array(freq_list) - freq0)/freq0/_MOM_COMPACT
            tunex_array = _np.array(tunex_list)
            tuney_array = _np.array(tuney_list)

            chromx, _ = _np.polyfit(den, tunex_array, deg=1, cov=True)
            chromy, _ = _np.polyfit(den, tuney_array, deg=1, cov=True)

            self._chrom_mon[0] = chromx[0]
            self._chrom_mon[1] = chromy[0]

            self.run_callbacks('MeasChromX-Mon', self._chrom_mon[0])
            self.run_callbacks('MeasChromY-Mon', self._chrom_mon[1])

            self._meas_chrom_status = _Const.MeasMon.Completed
            log_msg = 'Chromaticity measurement completed!'

        self.run_callbacks('MeasChromStatus-Mon', self._meas_chrom_status)
        self._measuring_chrom = False
        self.run_callbacks('Log-Mon', log_msg)
        return not aborted

    def _estimate_calc_chrom(self):
        sfam_deltasl = len(self._psfams)*[0]
        for fam_idx, fam in enumerate(self._psfams):
            sfam_deltasl[fam_idx] = \
                self._lastcalc_sl[fam] - self._psfam_nom_intstr[fam_idx]

        calc_estim = self._opticscorr.calculate_opticsparam(sfam_deltasl)
        self.run_callbacks('CalcChromX-Mon', calc_estim[0])
        self.run_callbacks('CalcChromY-Mon', calc_estim[1])

    # ---------- callbacks ----------

    def _callback_estimate_chrom(self, pvname, value, **kws):
        if value is None:
            return
        fam = _SiriusPVName(pvname).dev
        self._psfam_intstr_rb[fam] = value

        sfam_deltasl = len(self._psfams)*[0]
        for fam_idx, fam in enumerate(self._psfams):
            sfam_deltasl[fam_idx] = \
                self._psfam_intstr_rb[fam] - self._psfam_nom_intstr[fam_idx]

        self._optprm_est = self._opticscorr.calculate_opticsparam(sfam_deltasl)
        self.run_callbacks('ChromX-Mon', self._optprm_est[0])
        self.run_callbacks('ChromY-Mon', self._optprm_est[1])



"""Generic class to handle callbacks."""


class Callback:
    """Generic class to handle callbacks."""

    def __init__(self, callback=None):
        """Initialize Instance.

        callback can be any callable or list/tuple of callables.
        """
        self._callbacks = dict()
        if callback is None:
            return
        elif isinstance(callback, (tuple, list)):
            for thiscb in callback:
                self.add_callback(thiscb)
        else:
            self.add_callback(callback)

    @property
    def has_callbacks(self):
        """Return a flag indicating whether object has a callback attached.

        Returns:
            bool: whether or not there is a registered callback.

        """
        return bool(self._callbacks)

    def add_callback(self, callback, index=None, **kw):
        """Add a callback to a PV.

        Optional keyword arguments set here will be preserved and passed
        on to the callback at runtime.

        Note that there might be multiple callbacks, so that each
        has a unique index (small integer) that is returned by
        add_callback.  This index is needed to remove a callback.
        """
        if not callable(callback):
            raise ValueError("Tried to set non callable as a callback")
        if index is None:
            index = 0
            if self._callbacks:
                index = 1 + max(self._callbacks.keys())
        self._callbacks[index] = (callback, kw)
        return index

    def remove_callback(self, index=None):
        """Remove a callback by index."""
        if index in self._callbacks:
            self._callbacks.pop(index)

    def clear_callbacks(self):
        """Clear all callbacks."""
        self._callbacks = {}

    def run_callbacks(self, *args, **kwd):
        """Run all user-defined callbacks with the current data.

        Normally, this is to be run automatically on event, but
        it is provided here as a separate function for testing
        purposes.
        """
        for index in sorted(list(self._callbacks.keys())):
            self.run_callback(index, *args, **kwd)

    def run_callback(self, index, *args, **kwd):
        """Run a specific user-defined callback, specified by index.

        Note that callback functions are called with keyword/val
        arguments including:
            - any positional argument for the specific application
            - any keyword argument for the specific application
            - keyword args included in add_callback()
            - keyword 'cb_info' = (index, self)
        where the 'cb_info' is provided as a hook so that a callback
        function  that fails may de-register itself.
        """
        try:
            fcn, kwargs = self._callbacks[index]
        except KeyError:
            return
        kwargs2use = kwargs.copy()
        kwargs2use.update(kwd)
        kwargs2use['cb_info'] = (index, self)
        if callable(fcn):
            fcn(*args, **kwargs2use)



"""Synchronized Devices."""

from ..namesys import SiriusPVName as _SiriusPVName
from .device import Device as _Device


class DevicesSync(_Device):
    """Synchronized devices."""

    def __init__(
            self, devnames, propty_sync, propty_async=None,
            devname=None, auto_monitor_mon=False):
        """."""
        self._devnames = [_SiriusPVName(dev) for dev in devnames]
        self._props_sync = list(propty_sync)
        self._props_async = [] if propty_async is None else propty_async

        # get properties
        properties, self._prop2prop = self._get_properties()

        # call base class constructor
        super().__init__(
            devname, auto_monitor_mon=auto_monitor_mon, props2init=properties)

    @property
    def devnames(self):
        """Return device names."""
        return self._devnames

    @property
    def properties_sync(self):
        """Return sync properties."""
        return self._props_sync

    @property
    def properties_async(self):
        """Return async properties."""
        return self._props_async

    @property
    def synchronized(self):
        """Return True if devices are synchronized."""
        for devprop in self._props_sync:
            props = self._prop2prop[devprop]
            values = {self[prop] for prop in props}
            if len(values) > 1:
                return False
        return True

    def get_value(self, propty):
        """Return property value."""
        if not self.connected:
            return
        props = self._prop2prop[propty]
        values = [self[prop] for prop in props]
        if None in values:
            # NOTE: None values sometimes are being returned
            # even after connected check!
            # This is happening with 'TS-01:PU-EjeSeptG:Voltage-Mon'
            # of sirius-ioc-as-pu-conv, in particular.
            return
        return sum(values) / len(values)

    def set_value(self, propty, value):
        """Set property."""
        if not self.connected:
            return
        props = self._prop2prop[propty]
        for prop in props:
            self[prop] = value

    def _get_properties(self):
        properties = list()
        prop2prop = dict()
        for devname in self._devnames:
            for propty in self._props_sync + self._props_async:
                pvname = devname.substitute(propty=devname.propty_name+propty)
                if propty not in prop2prop:
                    prop2prop[propty] = [pvname]
                else:
                    prop2prop[propty].append(pvname)
                properties.append(pvname)
        return properties, prop2prop



"""Main Module of the IOC Logic."""

import time as _time
from datetime import datetime as _datetime
import logging as _log
from copy import deepcopy as _dcopy
from threading import Thread as _Thread

import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..callbacks import Callback as _Callback
from ..epics import SiriusPVTimeSerie as _SiriusPVTimeSerie, PV as _PV
from ..envars import VACA_PREFIX as _vaca_prefix
from ..clientarch import ClientArchiver as _ClientArch
from ..pwrsupply.csdev import Const as _PSc
from ..search import LLTimeSearch as _LLTimeSearch
from ..oscilloscope import Keysight as _Keysight, Scopes as _Scopes

from .csdev import Const as _Const, \
    get_currinfo_database as _get_database


class _CurrInfoApp(_Callback):
    """."""

    def __init__(self):
        """."""
        super().__init__()
        self._pvs_database = None

    def init_database(self):
        """."""

    @property
    def pvs_database(self):
        """."""
        return _dcopy(self._pvs_database)

    def process(self, interval):
        """."""
        _time.sleep(interval)

    def read(self, reason):
        """."""
        _ = reason
        return None

    def write(self, reason, value):
        """."""
        _, _ = reason, value
        return True

    def close(self):
        """."""

    @staticmethod
    def _get_value_from_arch(pvname):
        carch = _ClientArch()
        datetime = _datetime.now().isoformat() + '-03:00'
        return carch.getData(pvname, datetime, datetime)


class _ASCurrInfoApp(_CurrInfoApp):
    """."""

    INDICES1 = _get_namedtuple(
        'Indices1',
        ('NAME', 'CURR', 'STT', 'MIN', 'MAX', 'AVG', 'STD', 'COUNT'))
    # Some scopes does not return STT.
    INDICES2 = _get_namedtuple(
        'Indices2',
        ('NAME', 'CURR', 'MIN', 'MAX', 'AVG', 'STD', 'COUNT'))

    OSC = _Scopes.LI_DI_ICTOSC
    ACC = ''
    ICT1 = ''
    ICT2 = ''
    CHARGE_THRESHOLD = 0.05  # [nC]

    def __init__(self):
        super().__init__()
        self._pvs_database = _get_database(self.ACC)
        self._meas = None

        self.osc_obj = _Keysight(scope=self.OSC)

    def process(self, interval):
        """."""
        tini = _time.time()
        self._get_measurement()
        self._update_pvs(self.ACC, self.ICT1, self.ICT2)
        dtim = _time.time() - tini
        if dtim <= interval:
            _time.sleep(interval - dtim)
        else:
            _log.warning(f'IOC took {dtim*1000:.3f} ms in update loop.')

    def _get_measurement(self):
        try:
            self.osc_obj.connect()
            meas = self.osc_obj.send_command(b":MEASure:RESults?\n")
            self._meas = meas.split(',')
        except Exception as err:
            _log.error(str(err))
        finally:
            self.osc_obj.close()

    def _update_pvs(self, acc, ict1, ict2):
        """."""
        meas = self._meas
        if not meas:
            _log.warning('Measurement list is empty.')
            return

        # Check if measurement for each ICT has the length we expect:
        if not len(meas) % len(self.INDICES1):
            indcs = self.INDICES1
        elif not len(meas) % len(self.INDICES2):
            indcs = self.INDICES2
        else:
            _log.warning(
                'Measurement list size does not match required length.')
            return

        name = acc + '-ICT1'
        idxict1 = [i for i, val in enumerate(meas) if name in val]
        if not idxict1:
            _log.warning(f'Could not find data for {name}.')
            return
        idxict1 = idxict1.pop()

        name = acc + '-ICT2'
        idxict2 = [i for i, val in enumerate(meas) if name in val]
        if not idxict2:
            _log.warning(f'Could not find data for {name}.')
            return
        idxict2 = idxict2.pop()

        try:
            chg1 = float(meas[idxict1 + indcs.CURR]) * 1e9
            ave1 = float(meas[idxict1 + indcs.AVG]) * 1e9
            min1 = float(meas[idxict1 + indcs.MIN]) * 1e9
            max1 = float(meas[idxict1 + indcs.MAX]) * 1e9
            std1 = float(meas[idxict1 + indcs.STD]) * 1e9
            cnt1 = int(float(meas[idxict1 + indcs.COUNT]))
            chg2 = float(meas[idxict2 + indcs.CURR]) * 1e9
            ave2 = float(meas[idxict2 + indcs.AVG]) * 1e9
            min2 = float(meas[idxict2 + indcs.MIN]) * 1e9
            max2 = float(meas[idxict2 + indcs.MAX]) * 1e9
            std2 = float(meas[idxict2 + indcs.STD]) * 1e9
            cnt2 = int(float(meas[idxict2 + indcs.COUNT]))
        except IndexError:
            _log.warning('Problem reading data.')
            return

        eff = 0 if chg1 == 0 else min(100 * chg2/chg1, 110)
        effave = 0 if ave1 == 0 else 100 * ave2/ave1

        self.run_callbacks(ict1 + ':Charge-Mon', chg1)
        self.run_callbacks(ict1 + ':ChargeAvg-Mon', ave1)
        self.run_callbacks(ict1 + ':ChargeMin-Mon', min1)
        self.run_callbacks(ict1 + ':ChargeMax-Mon', max1)
        self.run_callbacks(ict1 + ':ChargeStd-Mon', std1)
        self.run_callbacks(ict1 + ':PulseCount-Mon', cnt1)
        self.run_callbacks(ict2 + ':Charge-Mon', chg2)
        self.run_callbacks(ict2 + ':ChargeAvg-Mon', ave2)
        self.run_callbacks(ict2 + ':ChargeMin-Mon', min2)
        self.run_callbacks(ict2 + ':ChargeMax-Mon', max2)
        self.run_callbacks(ict2 + ':ChargeStd-Mon', std2)
        self.run_callbacks(ict2 + ':PulseCount-Mon', cnt2)
        if chg1 <= self.CHARGE_THRESHOLD:
            return
        name = acc + '-Glob:AP-CurrInfo:'
        self.run_callbacks(name + 'TranspEff-Mon', eff)
        self.run_callbacks(name + 'TranspEffAvg-Mon', effave)


class TSCurrInfoApp(_ASCurrInfoApp):
    """."""

    OSC = _Scopes.AS_DI_FCTDIG
    ACC = 'TS'
    ICT1 = 'TS-01:DI-ICT'
    ICT2 = 'TS-04:DI-ICT'


class LICurrInfoApp(_ASCurrInfoApp):
    """Linac IOC will Also provide TB PVs."""

    OSC_IP = _Scopes.LI_DI_ICTOSC
    ACC = 'LI'
    LIICT1 = 'LI-01:DI-ICT-1'
    LIICT2 = 'LI-01:DI-ICT-2'
    TBICT1 = 'TB-02:DI-ICT'
    TBICT2 = 'TB-04:DI-ICT'

    def _update_pvs(self, *args):
        super()._update_pvs('LI', self.LIICT1, self.LIICT2)
        super()._update_pvs('TB', self.TBICT1, self.TBICT2)


class BOCurrInfoApp(_CurrInfoApp):
    """Main Class."""

    HARMNUM = 828
    INTCURR_INTVL = 53.5 * 1e-3 / 3600  # [h]
    MAX_CURRENT = 1.0  # [A]
    ENERGY2TIME = {  # energy: time[s]
        '150MeV': 0.0000,
        '1GeV': 0.0859,
        '2GeV': 0.1863,
        '3GeV': 0.2750}

    def __init__(self):
        """Class constructor."""
        super().__init__()

        # consts
        self._pvs_database = _get_database('BO')
        self._prefix = _vaca_prefix + ('-' if _vaca_prefix else '')
        self._dcct = 'BO-35D:DI-DCCT'

        # initialize vars
        self._samplecnt = None
        self._measperiod = None
        self._reliablemeas = None
        self._last_raw_reading = None

        currthold_db = self._pvs_database['CurrThold-SP']
        self._currthold = currthold_db['value']
        self._currthold_lolim = currthold_db['lolim']
        self._currthold_hilim = currthold_db['hilim']

        self._rampeff = None
        self._currents = dict()
        self._charges = dict()
        for k in self.ENERGY2TIME:
            # currents
            self._currents[k] = 0.0
            # charges
            ppty = 'Charge'+k+'-Mon'
            data = self._get_value_from_arch('BO-Glob:AP-CurrInfo:'+ppty)
            if data is None:
                charge = 0.0
            else:
                charge = data['value'][0]
            self._charges[k] = charge

        data = self._get_value_from_arch(
            'BO-Glob:AP-CurrInfo:IntCurrent3GeV-Mon')
        if data is None:
            self._intcurrent3gev = 0.0
        else:
            self._intcurrent3gev = data['value'][0]

        # PVs
        self._rawreadings_pv = _PV(
            self._prefix+self._dcct+':RawReadings-Mon',
            connection_timeout=0.05, callback=self._callback_get_rawreadings,
            auto_monitor=True)
        self._samplecnt_pv = _PV(
            self._prefix+self._dcct+':FastSampleCnt-RB',
            connection_timeout=0.05, callback=self._callback_get_samplecnt)
        self._measperiod_pv = _PV(
            self._prefix+self._dcct+':FastMeasPeriod-RB',
            connection_timeout=0.05, callback=self._callback_get_measperiod)
        self._reliablemeas_pv = _PV(
            self._prefix+self._dcct+':ReliableMeas-Mon',
            connection_timeout=0.05, callback=self._callback_get_reliablemeas)
        self._rffreq_pv = _PV(
            self._prefix+'RF-Gen:GeneralFreq-RB', connection_timeout=0.05)

    def init_database(self):
        """Set initial PV values."""
        for k in self.ENERGY2TIME:
            ppty = 'Charge'+k+'-Mon'
            self.run_callbacks(ppty, self._charges[k])
            ppty = 'Current'+k+'-Mon'
            self.run_callbacks(ppty, self._currents[k])
        self.run_callbacks('RawReadings-Mon', self._last_raw_reading)
        self.run_callbacks('IntCurrent3GeV-Mon', self._intcurrent3gev)

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        status = False
        if reason == 'CurrThold-SP':
            if self._currthold_lolim < value < self._currthold_hilim:
                self._currthold = value
                self.run_callbacks('CurrThold-RB', value)
                status = True
        return status

    # ----- callbacks -----

    def _callback_get_samplecnt(self, value, **kws):
        """Get SampleCnt-RB value."""
        _ = kws
        self._samplecnt = value

    def _callback_get_measperiod(self, value, **kws):
        """Get MeasPeriod-RB value."""
        _ = kws
        self._measperiod = value

    def _callback_get_reliablemeas(self, value, **kws):
        """Get ReliableMeas-Mon value."""
        _ = kws
        self._reliablemeas = value

    def _callback_get_rawreadings(self, value, **kws):
        _ = kws
        self._last_raw_reading = value
        self._update_pvs()
        self.run_callbacks('RawReadings-Mon', value)

    # ----- auxiliar methods -----

    def _update_pvs(self):
        """Update current, charges and ramp efficiency."""
        if self._samplecnt is None or self._measperiod is None:
            # ignore None values in initializations
            return
        if self._reliablemeas != 0:
            # do not calculate if dcct measure is not reliable
            return
        if len(self._last_raw_reading) != self._samplecnt:
            # ignore dcct measurements with wrong sample count
            return

        times = _np.linspace(0.0, self._measperiod, self._samplecnt)  # [ms]

        # calculate offset
        time_offset = self.ENERGY2TIME['3GeV'] + 0.0250
        idx_offset = _np.where(_np.isclose(times, time_offset, atol=0.0005))[0]
        try:
            samples_offset = self._last_raw_reading[idx_offset[0]:]
            offset = _np.mean(samples_offset)
        except (IndexError, TypeError):
            offset = 0.0

        # update pvs
        for energy, time in self.ENERGY2TIME.items():
            idx = _np.where(_np.isclose(times, time, atol=0.0005))[0]
            if idx.size:
                # currents
                current = self._last_raw_reading[idx[0]] - offset
                if current < self._currthold:
                    current = 0.0
                self._currents[energy] = current
                self.run_callbacks('Current'+str(energy)+'-Mon',
                                   self._currents[energy])
                # charges
                if current < self.MAX_CURRENT:
                    self._charges[energy] += \
                        1e6 * current * self.HARMNUM / self._rffreq_pv.value
                    if energy == '3GeV':
                        self._intcurrent3gev += \
                            current * self.INTCURR_INTVL  # [mA.h]
                        self.run_callbacks(
                            'IntCurrent3GeV-Mon', self._intcurrent3gev)
                else:
                    _log.warning(
                        'Current {0} value is too high: '
                        '{1:.3g}A.'.format(energy, current))
                self.run_callbacks('Charge'+str(energy)+'-Mon',
                                   self._charges[energy])

        c150mev = self._currents['150MeV']
        c3gev = self._currents['3GeV']

        # ramp efficiency
        if c150mev > c3gev:
            self._rampeff = 100*c3gev/c150mev
            self.run_callbacks('RampEff-Mon', self._rampeff)


class SICurrInfoApp(_CurrInfoApp):
    """Main Class."""

    HARMNUM = 864
    HARMNUM_RATIO = 864 / 828
    CURR_THRESHOLD = 0.02  # [mA]
    MAX_CURRENT = 1.0  # [A]

    def __init__(self):
        """Class constructor."""
        super().__init__()
        self._pvs_database = _get_database('SI')
        self._prefix = _vaca_prefix + ('-' if _vaca_prefix else '')

        # initialize vars
        self._time0 = _time.time()
        self._thread = None
        self._current_value = 0
        self._current_13c4_value = None
        self._current_14c4_value = None
        self._dcct_mode = _Const.DCCT.DCCT13C4
        self._dcctfltcheck_mode = _Const.DCCTFltCheck.Off
        self._reliablemeas_13c4_value = 0
        self._reliablemeas_14c4_value = 0
        self._storedebeam_value = 0
        self._storedebeam_13c4_value = 0
        self._storedebeam_14c4_value = 0
        self._is_cycling = False
        self._injeff = 0.0
        self._injcurr = 0.0
        self._injcharge = 0.0
        self._injcount = 0
        data = self._get_value_from_arch('SI-Glob:AP-CurrInfo:Charge-Mon')
        if data is None:
            self._charge = 0.0
        else:
            self._charge = data['value'][0]

        # pvs
        self._current_13c4_pv = _PV(
            self._prefix+'SI-13C4:DI-DCCT:Current-Mon',
            connection_timeout=0.05)
        self._current_14c4_pv = _PV(
            self._prefix+'SI-14C4:DI-DCCT:Current-Mon',
            connection_timeout=0.05)
        self._storedebeam_13c4_pv = _PV(
            self._prefix+'SI-13C4:DI-DCCT:StoredEBeam-Mon',
            connection_timeout=0.05,
            callback=self._callback_get_storedebeam)
        self._storedebeam_14c4_pv = _PV(
            self._prefix+'SI-14C4:DI-DCCT:StoredEBeam-Mon',
            connection_timeout=0.05,
            callback=self._callback_get_storedebeam)
        self._reliablemeas_13c4_pv = _PV(
            self._prefix+'SI-13C4:DI-DCCT:ReliableMeas-Mon',
            connection_timeout=0.05,
            callback=self._callback_get_reliablemeas)
        self._reliablemeas_14c4_pv = _PV(
            self._prefix+'SI-14C4:DI-DCCT:ReliableMeas-Mon',
            connection_timeout=0.05,
            callback=self._callback_get_reliablemeas)
        self._dipole_opmode_pv = _PV(
            self._prefix+'SI-Fam:PS-B1B2-1:OpMode-Sts',
            connection_timeout=0.05,
            callback=self._callback_get_dipole_opmode)
        self._bo_curr3gev_pv = _PV(
            self._prefix+'BO-Glob:AP-CurrInfo:Current3GeV-Mon',
            connection_timeout=0.05)
        self._ti_injcount_pv = _PV(
            self._prefix+_LLTimeSearch.get_evg_name()+':InjCount-Mon',
            connection_timeout=0.05,
            callback=self._callback_get_injcount)
        self._eg_trigps_pv = _PV(
            self._prefix+'LI-01:EG-TriggerPS:status',
            connection_timeout=0.05)
        self._rffreq_pv = _PV(
            self._prefix+'RF-Gen:GeneralFreq-RB', connection_timeout=0.05)

        self._current_13c4_buffer = _SiriusPVTimeSerie(
            pv=self._current_13c4_pv, time_window=0.5, use_pv_timestamp=False)
        self._current_14c4_buffer = _SiriusPVTimeSerie(
            pv=self._current_14c4_pv, time_window=0.5, use_pv_timestamp=False)

        self._current_13c4_pv.add_callback(self._callback_get_dcct_current)
        self._current_14c4_pv.add_callback(self._callback_get_dcct_current)
        self._bo_curr3gev_pv.add_callback(self._callback_get_bo_curr3gev)

    def init_database(self):
        """Set initial PV values."""
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:StoredEBeam-Mon', self._storedebeam_value)
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:Current-Mon', self._current_value)
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:InjEff-Mon', self._injeff)
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:Charge-Mon', self._charge)
        self.run_callbacks(
            'AS-Glob:AP-CurrInfo:InjCount-Mon', self._injcount)
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:InjCurr-Mon', self._injcurr)
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:InjCharge-Mon', self._injcharge)

    def read(self, reason):
        """Read from IOC database."""
        value = None
        if 'Charge' in reason:
            timestamp = _time.time()
            if self._storedebeam_value and not self._is_cycling:
                dtm = (timestamp - self._time0)  # Delta t [s]
                current = self._current_value/1000  # Current [A]
                if current < self.MAX_CURRENT:
                    inc_charge = current*dtm/3600  # Charge [A.h]
                    self._charge += inc_charge
                    self.run_callbacks(
                        'SI-Glob:AP-CurrInfo:Charge-Mon', self._charge)
                else:
                    _log.warning(
                        'Current value is too high: {0:.3f}A.'.format(current))
                value = self._charge
            self._time0 = timestamp
        return value

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        status = False
        if reason == 'SI-Glob:AP-CurrInfo:DCCT-Sel':
            if self._dcctfltcheck_mode == _Const.DCCTFltCheck.Off:
                done = self._update_dcct_mode(value)
                if done:
                    self.run_callbacks(
                        'SI-Glob:AP-CurrInfo:DCCT-Sts', self._dcct_mode)
                    status = True
        elif reason == 'SI-Glob:AP-CurrInfo:DCCTFltCheck-Sel':
            self._update_dcctfltcheck_mode(value)
            self.run_callbacks(
                'SI-Glob:AP-CurrInfo:DCCTFltCheck-Sts',
                self._dcctfltcheck_mode)
            status = True
        return status

    # ----- handle writes -----
    def _update_dcct_mode(self, value):
        if self._dcct_mode == value:
            return False
        if value in [_Const.DCCT.DCCT13C4, _Const.DCCT.DCCT14C4]:
            self._dcct_mode = value
            return True
        return False

    def _update_dcct_mode_by_relblmeas(self):
        if self._reliablemeas_13c4_value == 0:
            mode = _Const.DCCT.DCCT13C4
        elif self._reliablemeas_14c4_value == 0:
            mode = _Const.DCCT.DCCT14C4
        else:
            mode = self._dcct_mode
        if mode != self._dcct_mode:
            self._dcct_mode = mode
            self.run_callbacks('SI-Glob:AP-CurrInfo:DCCT-Sts', self._dcct_mode)

    def _update_dcctfltcheck_mode(self, value):
        if self._dcctfltcheck_mode != value:
            if value == _Const.DCCTFltCheck.On:
                self._update_dcct_mode_by_relblmeas()
            self._dcctfltcheck_mode = value

    # ----- callbacks -----
    def _callback_get_dcct_current(self, pvname, value, **kws):
        _ = kws
        if '13C4' in pvname:
            self._current_13c4_value = value
            self._current_13c4_buffer.acquire()
        elif '14C4' in pvname:
            self._current_14c4_value = value
            self._current_14c4_buffer.acquire()

        # update current and charge
        current = self._get_current()
        if current is None:
            return
        self._current_value = current

        # update pvs
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:Current-Mon', self._current_value)

    def _callback_get_storedebeam(self, pvname, value, **kws):
        _ = kws
        if '13C4' in pvname:
            self._storedebeam_13c4_value = value
        elif '14C4' in pvname:
            self._storedebeam_14c4_value = value

        if self._dcct_mode == _Const.DCCT.DCCT13C4:
            self._storedebeam_value = self._storedebeam_13c4_value
        elif self._dcct_mode == _Const.DCCT.DCCT14C4:
            self._storedebeam_value = self._storedebeam_14c4_value
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:StoredEBeam-Mon', self._storedebeam_value)

    def _callback_get_reliablemeas(self, pvname, value, **kws):
        _ = kws
        if '13C4' in pvname:
            self._reliablemeas_13c4_value = value
        elif '14C4' in pvname:
            self._reliablemeas_14c4_value = value

        if self._dcctfltcheck_mode == _Const.DCCTFltCheck.On:
            self._update_dcct_mode_by_relblmeas()

    def _callback_get_dipole_opmode(self, value, **kws):
        _ = kws
        self._is_cycling = bool(value == _PSc.States.Cycle)

    def _callback_get_bo_curr3gev(self, value, timestamp, **kws):
        """Get BO Current3GeV-Mon and update InjEff PV."""
        _ = value
        _ = kws
        _ = timestamp
        self._thread = _Thread(target=self._update_injeff, daemon=True)
        self._thread.start()

    def _callback_get_injcount(self, value, **kws):
        _ = kws
        if value == 0:
            self._injcount = 0
        else:
            if self._eg_trigps_pv.value:
                self._injcount += 1
        self.run_callbacks('AS-Glob:AP-CurrInfo:InjCount-Mon', self._injcount)

    def _update_injeff(self):
        # Sleep some time here to ensure SI DCCT will have been updated
        _time.sleep(0.21)

        # get booster current
        bo_curr = self._bo_curr3gev_pv.value

        # choose current PV
        if self._dcct_mode == _Const.DCCT.DCCT13C4:
            buffer = self._current_13c4_buffer
        else:
            buffer = self._current_14c4_buffer
        timestamp_dq, value_dq = buffer.serie
        timestamp_dq = _np.asarray(timestamp_dq)
        value_dq = _np.asarray(value_dq)

        self._injcurr = 0.0
        self._injeff = 0.0
        # check if buffer not empty and if there is valid current in Booster
        if timestamp_dq.size and bo_curr >= self.CURR_THRESHOLD:
            self._injcurr = value_dq.ptp()  # mA
            self._injeff = 100*(self._injcurr/bo_curr) * self.HARMNUM_RATIO

        # calculate injected charge: 1e6 * mA / Hz = nC
        self._injcharge = 1e6*self._injcurr*self.HARMNUM/self._rffreq_pv.value

        # update pvs
        self.run_callbacks('SI-Glob:AP-CurrInfo:InjEff-Mon', self._injeff)
        self.run_callbacks('SI-Glob:AP-CurrInfo:InjCurr-Mon', self._injcurr)
        self.run_callbacks(
            'SI-Glob:AP-CurrInfo:InjCharge-Mon', self._injcharge)

    # ----- auxiliar methods -----

    def _get_current(self):
        if self._dcct_mode == _Const.DCCT.DCCT13C4:
            current = self._current_13c4_value
        elif self._dcct_mode == _Const.DCCT.DCCT14C4:
            current = self._current_14c4_value
        return current



"""Keysight oscilloscope classes."""

import socket as _socket
import time as _time
import sys as _sys
import numpy as _np

from . import scopes as _scopes


class Keysight:
    """Keysight Oscilloscopes.

    For users' and programmers' guides (model DSOS104A), see
    http://10.0.38.42/control-system-constants/documentation/keysight/
    for original source:
        https://www.keysight.com/us/en/support/DSOS104A/
        high-definition-oscilloscope-1-ghz-4-analog-channels.html.


    Example
    =======

    from siriuspy.oscilloscope import Keysight, ScopeSignals
    import matplotlib.pyplot as plt

    scope = Keysight(scopesignal=ScopeSignals.SI_FILL_PATTERN)
    print(scope.scope_name)
    wavet, waved = scope.wfm_get_data()
    plt.plot(wavet, waved)
    plt.show()

    """

    SOCKET_TIMEOUT = 10  # [s]

    def __init__(self, scope=None, scopesignal=None):
        """."""
        if scopesignal and isinstance(scopesignal, tuple):
            self.host = scopesignal[0]
            self.port = scopesignal[1]
            self.chan = scopesignal[2]
        elif scope and isinstance(scope, str):
            self.host = scope
            self.port = 5025  # keysight default?
            self.chan = 'CHAN1'
        else:
            raise NotImplementedError
        self._socket = None

    @property
    def scope_name(self):
        """."""
        return _scopes.ScopeSignals.get_scope_name(self.host)

    def connect(self):
        """."""
        self._socket = _socket.socket(
            _socket.AF_INET,  # Internet
            _socket.SOCK_STREAM)  # TCP
        self._socket.settimeout(Keysight.SOCKET_TIMEOUT)
        self._socket.connect((self.host, self.port))

    def close(self):
        """."""
        self._socket.close()

    def wfm_enable(self):
        """Enable scope waveform acquisition."""
        return self.send_command(b'*IDN?\r\n')

    def wfm_config(self, wait_trigger=False):
        """Set scope waveform format."""
        self.send_command(b":WAVeform:FORMat WORD\n", get_res=False)
        dataformat = self.send_command(b":WAVeform:FORMat?\n")
        print('Data format:', dataformat)
        # Set bit order to MSB First
        self.send_command(b":WAVeform:BYTeorder MSBF\n", get_res=False)
        # Acquire
        if wait_trigger:
            self.send_command(b':DIG\n', get_res=False)
        else:
            self.send_command(b':RUN\n', get_res=False)

    def wfm_acquire(self, channel):
        """Acquire scope waveform."""
        if isinstance(channel, tuple):
            channel = channel[2]

        # Get the number of waveform points
        points = self.send_command(b":WAVeform:POINts?\n")
        print('Points:', points)

        # Get sample rate
        srate = self.send_command(b":ACQuire:SRATe?\n")
        print('Sample rate:', srate)

        # Get bandwidth
        bdw = self.send_command(b":ACQuire:BANDwidth:FRAMe?\n")
        print('Bandwidth:', bdw)

        # Set the waveform channel source
        self.send_command(
            b":WAVeform:SOURce " + channel.encode('ascii')+b"\n",
            get_res=False)
        channel = self.send_command(b":WAVeform:SOURce?\n")

        # Get scales
        xinc = self.send_command(b":WAVeform:XINCrement?\n")
        xinc = float(xinc)
        print('Horizontal Scale:', xinc)
        yinc = self.send_command(b":WAVeform:YINCrement?\n")
        yor = self.send_command(b":WAVeform:YORigin?\n")
        yinc = float(yinc)
        yor = float(yor)
        print('Vertical Scale:', xinc, yor)

        # Data aquisition
        self.send_command(b":WAVeform:STReaming OFF\n", get_res=False)
        self.send_command(b":WAVeform:DATA?\n", get_res=False)
        _ = self._socket.recv(1)  # ignore marker '#'
        num = int(self._socket.recv(1).decode('ascii'))
        datanum = int(self._socket.recv(num).decode('ascii'))
        dataraw = b''

        # Return scope to RUN mode
        self.send_command(b":RUN\n", get_res=False)
        while len(dataraw) < datanum:
            dataraw = dataraw + self._socket.recv(datanum)
        dataraw = dataraw[0:-1]  # remove EOF char

        va1 = _np.array(list(dataraw)[0::2])
        va0 = _np.array(list(dataraw)[1::2])
        va1 = va1[:va0.size]

        datay = ((va1 << 8) + va0 - 2**16*(va1 >> 7)) * yinc + yor

        datax = _np.arange(datay.size)*xinc
        return datax, datay, srate, bdw

    def wfm_get_data(self, channel=None, wait_trigger=False):
        """Enable and get sccope waveform data."""
        channel = channel or self.chan
        self.connect()
        wavet = None
        waved = None
        try:
            self.wfm_enable()
            self.wfm_config(wait_trigger)
            tini = _time.time()
            print('Acquiring ' + self.chan)
            wavet, waved, srate1, bdw1 = self.wfm_acquire(channel)
            print('Total acquisition time:', _time.time() - tini)
            # self.send_command(b":WAVeform:STReaming ON\n", get_res=False)
        except Exception:
            print('Close connection by exception')
            raise

        finally:
            self.close()

        return wavet, waved

    def stats_enable(self):
        """Enable scope measurement statistics info."""
        # Set bit order to MSB First
        self.send_command(b":MEASure:STATistics ON\n", get_res=False)
        self.send_command(b":MEASure:SENDvalid ON\n", get_res=False)

    def stats_acquire(self):
        """Return a dictionary of scope measurement statistics."""
        meas = self.send_command(b":MEASure:RESults?\n")
        data = Keysight.process_stats_meas(meas)
        return data

    def stats_get_data(self):
        """."""
        self.connect()
        try:
            self.stats_enable()
            tini = _time.time()
            print('Acquiring ' + self.chan)
            data = self.stats_acquire()
            print('Total acquisition time:', _time.time() - tini)
        except Exception:
            print('Close connection by exception')
            raise
        finally:
            self.close()
        return data

    def send_command(self, cmd, get_res=True):
        """."""
        self._socket.sendall(cmd)
        if get_res:
            return self._socket.recv(1024).decode('ascii')
        return

    @staticmethod
    def process_stats_meas(meas):
        """Process stats measurents."""
        # TODO: generalize or instantiate for each scope type
        meas = meas.split(',')
        data = dict()
        for i in range(0, len(meas) - len(meas) % 8, 8):
            datum = dict()
            label = meas[i]
            datum['current'] = float(meas[i+1])
            datum['state'] = int(float(meas[i+2]))
            datum['min'] = float(meas[i+3])
            datum['max'] = float(meas[i+4])
            datum['mean'] = float(meas[i+5])
            datum['std_dev'] = float(meas[i+6])
            datum['num_of_meas'] = int(float(meas[i+7]))
            data[label] = datum
        return data



"""Set important variables for the proper work of the package."""
import os as _os


# --- directories ---

DIR_ROOT = _os.environ.get(
    'ROOT_GROUP', default='/home')
DIR_EPICS_BASE = _os.environ.get(
    'EPICS_BASE', default='/usr/lib/epics')

DIR_SIRIUS = _os.environ.get(
    'LNLS_SIRIUS', default=_os.path.join(DIR_ROOT, 'sirius'))
DIR_FACS = _os.environ.get(
    'LNLS_FAC', default=_os.path.join(DIR_ROOT, 'facs'))
DIR_IMAS = _os.environ.get(
    'LNLS_IMA', default=_os.path.join(DIR_ROOT, 'imas'))

DIR_SIRIUS_CODE = _os.environ.get(
    'LNLS_SIRIUS', default=_os.path.join(DIR_ROOT, DIR_SIRIUS, 'repos'))
DIR_FACS_CODE = _os.path.join(
    DIR_ROOT, DIR_FACS, 'repos')

DIR_SIRIUS_CODE_CSCNSTS = _os.path.join(
    DIR_SIRIUS_CODE, 'control-system-constants')
DIR_SIRIUS_CODE_SIRIUSPY = _os.path.join(
    DIR_SIRIUS_CODE, 'dev-packages')
DIR_SIRIUS_CODE_HLA = _os.path.join(
    DIR_SIRIUS, 'hla')


# --- support applications IPs/URLs ---

SRVURL_RBACAUTH = _os.environ.get(
    'SIRIUS_URL_RBAC_AUTH', default='https://sirius-rbac-auth.lnls.br')
SRVURL_RBAC = _os.environ.get(
    'SIRIUS_URL_RBAC', default='https://rbac:8445')
SRVURL_NS = _os.environ.get(
    'SIRIUS_URL_NS', default='http://naming-service-wildfly:8089/names')
SRVURL_CCDB = _os.environ.get(
    'SIRIUS_URL_CCDB', default='http://ccdb:8083')
SRVURL_CABLES = _os.environ.get(
    'SIRIUS_URL_CABLES', default='http://cables:8086')
SRVURL_CSCONSTS = _os.environ.get(
    'SIRIUS_URL_CONSTS',
    default='http://10.0.38.59/control-system-constants')
SRVURL_CSCONSTS_2 = _os.environ.get(
    'SIRIUS_URL_CONSTS_2',
    default='http://10.128.255.3/control-system-constants')
SRVURL_LOGBOOK = _os.environ.get(
    'SIRIUS_URL_LOGBOOK',
    default='http://10.0.38.42/Olog')
SRVURL_CONFIGDB = _os.environ.get(
    'SIRIUS_URL_CONFIGDB',
    default='http://10.0.38.59/config-db')
SRVURL_CONFIGDB_2 = _os.environ.get(
    'SIRIUS_URL_CONFIGDB_2',
    default='http://10.128.255.3/config-db')
SRVURL_ARCHIVER = _os.environ.get(
    'SIRIUS_URL_ARCHIVER',
    default='https://10.0.38.42')


VACA_PREFIX = _os.environ.get('VACA_PREFIX', default='')



"""Energy Devices."""

from .syncd import DevicesSync as _DevicesSync


class Energy(_DevicesSync):
    """Energy devices."""

    class DEVICES:
        """."""

        TB = 'TB'
        BO = 'BO'
        TS = 'TS'
        SI = 'SI'
        ALL = (TB, BO, TS, SI)

    _props_sync = ('Energy-SP', 'Energy-RB', 'EnergyRef-Mon')
    _props_async = ('Energy-Mon', )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in Energy.DEVICES.ALL:
            raise NotImplementedError(devname)

        # get dipole names
        psnames_dip = self._get_dipole_devnames(devname)

        # call base class constructor
        super().__init__(
            psnames_dip,
            Energy._props_sync, Energy._props_async, devname)

    @property
    def energy(self):
        """Return Ref-Mon energy."""
        return self.get_value('EnergyRef-Mon')

    @energy.setter
    def energy(self, value):
        """Set energy."""
        self.set_value('Energy-SP', value)

    @property
    def energy_sp(self):
        """Return -SP energy."""
        return self.get_value('Energy-SP')

    @property
    def energy_mon(self):
        """Return -Mon energy."""
        return self.get_value('Energy-Mon')

    @staticmethod
    def _get_dipole_devnames(devname):
        if devname == Energy.DEVICES.TB:
            psnames_dipoles = ('TB-Fam:PS-B', )
        elif devname == Energy.DEVICES.BO:
            psnames_dipoles = ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2')
        elif devname == Energy.DEVICES.TS:
            psnames_dipoles = ('TS-Fam:PS-B', )
        elif devname == Energy.DEVICES.SI:
            psnames_dipoles = ('SI-Fam:PS-B1B2-1', 'SI-Fam:PS-B1B2-2')
        return psnames_dipoles



"""Reconstruction factories."""

from copy import deepcopy as _dcopy
import numpy as _np
from scipy.optimize import least_squares

from ..epics import PV as _PV
from ..search import PSSearch as _PSSearch, LLTimeSearch as _LLTimeSearch, \
    MASearch as _MASearch
from .ramp import BoosterRamp as _BORamp
from .waveform import Waveform as _Waveform, WaveformDipole as _WaveformDipole
from .magnet import get_magnet as _get_magnet


TIMEOUT_CONN = 0.05


class BODipRampFactory:
    """Class to rebuild Dipole energies from machine state."""

    _PSNAME_DIPOLES = ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2')
    _PSNAME_DIPOLE_REF = _PSNAME_DIPOLES[0]
    _ppties_2_fit = [
        'start_value',
        'rampup1_start_value',
        'rampup2_start_value',
        'rampdown_start_value',
        'rampup_smooth_value',
        'rampdown_stop_value',
        'rampdown_smooth_value',
    ]
    _ppties_2_use = [
        'duration',
        'rampup1_start_time',
        'rampup2_start_time',
        'rampdown_start_time',
        'rampdown_stop_time',
        'rampup_smooth_intvl',
        'rampdown_smooth_intvl',
    ]
    _PVs = dict()

    def __init__(self, ramp_config, waveform=None):
        """Init."""
        self._ramp_config = ramp_config
        self._duration = self._ramp_config.ps_ramp_duration
        self._magnet = _get_magnet(_MASearch.conv_psname_2_psmaname(
            BODipRampFactory._PSNAME_DIPOLE_REF))
        self._nrpoints = None
        self._times = None
        self.waveform = _np.array(waveform)
        self._create_pvs()

    @property
    def waveform(self):
        """Return waveform in current values."""
        return self._waveform

    @waveform.setter
    def waveform(self, waveform=None):
        """Set waveform in current values."""
        if waveform is None:
            waveform = _np.array([])
        self._waveform = waveform
        self._nrpoints = self._waveform.size
        self._times = _np.linspace(0, self._duration, self._nrpoints)

    def read_waveform(self):
        """Read waveform in current values from PVs."""
        if not BODipRampFactory._PVs:
            self._create_pvs()

        dipname = BODipRampFactory._PSNAME_DIPOLE_REF
        pvobj = BODipRampFactory._PVs[dipname]
        pvobj.wait_for_connection(10*TIMEOUT_CONN)
        if not pvobj.connected:
            raise ConnectionError('Dipole waveform PV is disconnected!')
        self._waveform = pvobj.get()
        self._nrpoints = self._waveform.size
        self._times = _np.linspace(0, self._duration, self._nrpoints)

    @property
    def dip_params(self):
        """Return dipole parameters rebuilt from PV values."""
        return self._generate_dip_params()

    # ----- private methods -----

    def _create_pvs(self):
        dipname = BODipRampFactory._PSNAME_DIPOLE_REF
        pvs = {dipname: _PV(
            dipname + ':Wfm-SP', connection_timeout=TIMEOUT_CONN)}
        BODipRampFactory._PVs = pvs

    def _generate_dip_params(self):
        if self._waveform is None:
            raise ValueError(
                'Waveform has None value. '
                'Set a valid value to waveform property '
                'or call read_waveform method.')

        init_params = self._get_initial_params()
        _x_init = [init_params[ppty] for ppty in
                   BODipRampFactory._ppties_2_fit]
        _x_final = least_squares(
            self._err_func, _x_init, method='lm').x
        final_params = {ppty: _x_final[i] for i, ppty in
                        enumerate(BODipRampFactory._ppties_2_fit)}

        result = dict()
        for ppty, value in final_params.items():
            energy_ppty = ppty.replace('value', 'energy')
            result[energy_ppty] = self._magnet.conv_current_2_strength(value)
        return result

    def _get_initial_params(self):
        params = dict()
        for ppty in BODipRampFactory._ppties_2_fit:
            if ppty in ['rampup2_start_value', 'rampdown_start_value',
                        'rampup_smooth_value', 'rampdown_smooth_value']:
                continue
            else:
                time_ppty_name = 'ps_ramp_'+ppty.replace('value', 'time')
                time = 0.0 if ppty == 'start_value' else getattr(
                    self._ramp_config, time_ppty_name)
                params[ppty] = _np.interp(time, self._times, self.waveform)

        def calc_intersection(s1_times, s2_times, time):
            slopes, offsets = list(), list()
            for s_times in [s1_times, s2_times]:
                ids = _np.arange(len(self._times))
                s_ids = _np.round(_np.interp(s_times, self._times, ids))
                slc = slice(int(s_ids[0]), int(s_ids[1]))
                s_x, s_y = self._times[slc], self.waveform[slc]
                s_off, s_slp = _np.polynomial.polynomial.polyfit(s_x, s_y, 1)
                offsets.append(s_off)
                slopes.append(s_slp)
            time_inter = -1 * _np.diff(offsets)/_np.diff(slopes)
            if not s1_times[-1] < time_inter < s2_times[0]:
                time_inter = time
            else:
                time_inter = time_inter[0]
            val_inter = slopes[0]*time_inter + offsets[0]
            return val_inter

        # calculate value for rampup2_start and rampdown_start
        ru1_times = [self._ramp_config.ps_ramp_rampup1_start_time,
                     (self._ramp_config.ps_ramp_rampup2_start_time -
                      self._ramp_config.ps_ramp_rampup_smooth_intvl/2)]
        ru2_times = [(self._ramp_config.ps_ramp_rampup2_start_time +
                      self._ramp_config.ps_ramp_rampup_smooth_intvl/2),
                     (self._ramp_config.ps_ramp_rampdown_start_time -
                      self._ramp_config.ps_ramp_rampdown_smooth_intvl/2)]
        rd_times = [(self._ramp_config.ps_ramp_rampdown_start_time +
                     self._ramp_config.ps_ramp_rampdown_smooth_intvl/2),
                    self._ramp_config.ps_ramp_rampdown_stop_time]

        ru_time = self._ramp_config.ps_ramp_rampup2_start_time
        params['rampup2_start_value'] = calc_intersection(
            ru1_times, ru2_times, ru_time)
        params['rampup_smooth_value'] = params['rampup2_start_value'] - \
            _np.interp(ru_time, self._times, self.waveform)

        rd_time = self._ramp_config.ps_ramp_rampdown_start_time
        params['rampdown_start_value'] = calc_intersection(
            ru2_times, rd_times, rd_time)
        params['rampdown_smooth_value'] = params['rampdown_start_value'] - \
            _np.interp(rd_time, self._times, self.waveform)

        return params

    def _err_func(self, params):
        new = _WaveformDipole()
        for ppty in BODipRampFactory._ppties_2_use:
            val = getattr(self._ramp_config, 'ps_ramp_' + ppty)
            setattr(new, ppty, val)
        for i, ppty in enumerate(BODipRampFactory._ppties_2_fit):
            setattr(new, ppty, params[i])
        error = self.waveform - new.currents
        return error


class BONormListFactory:
    """Class to rebuild Norm. Config. List from machine state."""

    _PSNAME_DIPOLES = ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2')
    _PSNAME_DIPOLE_REF = _PSNAME_DIPOLES[0]
    _PVs = dict()

    # desired reconstruction precision
    _DESIRED_PRECISION = 1e-5

    # thersholds to norm. configs time detection
    _THRESHOLD_DEFAULT = 1e-10
    _THRESHOLD_PROBLEM = 1e-4

    # loss factors consider that a change of 1e-1urad in a corrector is
    # comparable to a change of 1e-4 in tune fraction and 1e-2 in chromaticity
    _LOSS_FACTOR_CORRS = 1
    _LOSS_FACTOR_QUADS = 2e-5
    _LOSS_FACTOR_SEXTS = 1e-2

    # if considering only beam interval
    _BEAM_INTERVAL = [0.0, 300.0]

    def __init__(self, ramp_config, waveforms=None, opt_metric='strength',
                 opt_global=False, opt_times=False, use_config_times=False,
                 use_straigth_estim=True, consider_beam_interval=False):
        """Init."""
        if waveforms is None:
            waveforms = dict()
        self._opt_metric = opt_metric
        self._opt_global = opt_global
        self._opt_times = opt_times
        self._use_rampconfig_times = use_config_times
        self._use_straigth_estim = use_straigth_estim
        self._consider_beam_interval = consider_beam_interval

        # declaration of attributes (as pylint requires)
        self._wfms_strength = None
        self._waveforms = None

        self._ramp_config = ramp_config
        self._dipole = ramp_config.ps_waveform_get(self._PSNAME_DIPOLE_REF)
        self._duration = ramp_config.ps_ramp_duration
        self._nrpoints_fams = ramp_config.ps_ramp_wfm_nrpoints_fams
        self._times_fams = self._dipole.times
        self._dipstrgs_fams = self._dipole.strengths
        self._nrpoints_corrs = ramp_config.ps_ramp_wfm_nrpoints_corrs
        self._times_corrs = _np.linspace(
            0.0, self._duration, self._nrpoints_corrs)
        self._dipstrgs_corrs = _np.interp(
            self._times_corrs, self._dipole.times, self._dipole.strengths)

        self._psnames = _PSSearch.get_psnames({'sec': 'BO', 'dis': 'PS'})
        for dip in self._PSNAME_DIPOLES:
            self._psnames.remove(dip)

        self._wfms_current = waveforms
        self._norm_configs_dict = dict()
        if waveforms:
            self._create_waveform_objects()
            self._generate_nconf_dict()

    @property
    def psnames(self):
        """."""
        return self._psnames

    @property
    def waveforms(self):
        """Return a dict of psname: waveform in current values."""
        return self._wfms_current

    @waveforms.setter
    def waveforms(self, waveforms=None):
        """Set a dict of psname: waveform in current values."""
        if waveforms is None:
            waveforms = dict()
        self._wfms_current = waveforms
        if waveforms:
            self._create_waveform_objects()
            self._generate_nconf_dict()

    def read_waveforms(self):
        """Read waveform in current values from PVs."""
        if not BONormListFactory._PVs:
            self._create_pvs()

        self._waveforms = dict()
        for psname in self.psnames:
            pvobj = BONormListFactory._PVs[psname]
            pvobj.wait_for_connection(10*TIMEOUT_CONN)
            if not pvobj.connected:
                raise ConnectionError('There are disconnected PVs!')
            self._wfms_current[psname] = pvobj.get()

        self._create_waveform_objects()
        self._generate_nconf_dict()

    @property
    def normalized_configs(self):
        """Return dict of [time: normalized configuration]."""
        return _dcopy(self._norm_configs_dict)

    @property
    def desired_reconstr_precision(self):
        """Desired reconstruction precision."""
        return BONormListFactory._DESIRED_PRECISION

    @property
    def precision_reached(self):
        """Precision reached.

        Return if the reconstructed normalized configuration dict
        reached the desired  precision.
        """
        if not self._norm_configs_dict:
            return False

        r_new = _BORamp()
        attrs = [
            'ps_ramp_wfm_nrpoints_fams',
            'ps_ramp_wfm_nrpoints_corrs',
            'ps_ramp_duration',
            'ps_ramp_start_energy',
            'ps_ramp_rampup1_start_time',
            'ps_ramp_rampup1_start_energy',
            'ps_ramp_rampup2_start_time',
            'ps_ramp_rampup2_start_energy',
            'ps_ramp_rampup_smooth_intvl',
            'ps_ramp_rampup_smooth_energy',
            'ps_ramp_rampdown_start_time',
            'ps_ramp_rampdown_start_energy',
            'ps_ramp_rampdown_stop_time',
            'ps_ramp_rampdown_stop_energy',
            'ps_ramp_rampdown_smooth_intvl',
            'ps_ramp_rampdown_smooth_energy',
            'rf_ramp_rampup_start_time',
            'rf_ramp_rampup_stop_time',
            'rf_ramp_rampdown_start_time',
            'rf_ramp_rampdown_stop_time',
            'rf_ramp_bottom_voltage',
            'rf_ramp_top_voltage',
            'rf_ramp_bottom_phase',
            'rf_ramp_top_phase',
            'ti_params_injection_time',
            'ti_params_ejection_time',
            'ti_params_ps_ramp_delay',
            'ti_params_rf_ramp_delay',
        ]
        for attr in attrs:
            setattr(r_new, attr, getattr(self._ramp_config, attr))
        r_new.ps_normalized_configs_set(self.normalized_configs)

        max_error = 0.0
        for psname in self.psnames:
            error = _np.abs(
                self._ps2wfms[psname].currents -
                r_new.ps_waveform_get_currents(psname))
            max_error = max(max_error, max(error))
        if max_error < self.desired_reconstr_precision:
            return True, max_error
        return False, max_error

    def calc_psname_current_error(self, psname, nc_times, nc_strengths):
        """Calculate error vector in current.

        Calculate reconstruction error vector in current units for
        psname given normalized configurations times and strengths."""
        if 'Fam' in psname:
            wfm_times = self._times_fams
            dip_strgs = self._dipstrgs_fams
        else:
            wfm_times = self._times_corrs
            dip_strgs = self._dipstrgs_corrs
        wfm_strengths = _np.interp(wfm_times, nc_times, nc_strengths)
        magnet = _get_magnet(_MASearch.conv_psname_2_psmaname(psname))
        wfm_currents = magnet.conv_strength_2_current(
            strengths=wfm_strengths, strengths_dipole=dip_strgs)
        error = self._wfms_current[psname] - wfm_currents
        if self._consider_beam_interval:
            ini, end = _np.round(_np.interp(
                BONormListFactory._BEAM_INTERVAL,
                wfm_times, _np.arange(len(wfm_times))))
            error = error[int(ini):int(end)]
        return error

    def calc_psname_strength_error(self, psname, nc_times, nc_strengths):
        """Calculate error vector in strength.

        Calculate reconstruction error vector in strengths units for
        psname given normalized configurations times and strengths."""
        wfm_times = self._times_fams if 'Fam' in psname else self._times_corrs
        wfm_strengths = _np.interp(wfm_times, nc_times, nc_strengths)
        error = self._wfms_strength[psname] - wfm_strengths
        if self._consider_beam_interval:
            ini, end = _np.round(_np.interp(
                BONormListFactory._BEAM_INTERVAL,
                wfm_times, _np.arange(len(wfm_times))))
            error = error[int(ini):int(end)]
        return error

    # ----- private methods -----

    def _create_pvs(self):
        pvs = dict()
        for psname in self.psnames:
            pvs[psname] = _PV(
                psname + ':Wfm-SP', connection_timeout=TIMEOUT_CONN)
        BONormListFactory._PVs = pvs

    def _create_waveform_objects(self):
        self._ps2wfms = dict()
        self._wfms_strength = dict()
        for psname, wfm_curr in self._wfms_current.items():
            wfm_nrpoints = self._nrpoints_fams if 'Fam' in psname \
                else self._nrpoints_corrs
            self._ps2wfms[psname] = _Waveform(
                psname, self._dipole, currents=wfm_curr,
                wfm_nrpoints=wfm_nrpoints)
            self._wfms_strength[psname] = self._ps2wfms[psname].strengths

    def _calc_nconf_times(self, times, wfm, thres):
        if self._use_rampconfig_times:
            nconf_times = sorted(self._ramp_config.ps_normalized_configs_times)
            nconf_indcs = list()
            for tim in nconf_times:
                atol = self._duration/wfm.size
                idx = _np.where(_np.isclose(times, tim, atol=atol))[0]
                nconf_indcs.append(idx[-1])
            nconf_times = _np.array(nconf_times)
            nconf_indcs = _np.array(nconf_indcs)

            idd_bef = _np.r_[nconf_indcs, times.size-3]
            idd_aft = _np.r_[0, nconf_indcs]
        else:
            # calculate 1 e 2 derivatives
            diff1 = _np.diff(wfm)
            diff2 = _np.diff(diff1)

            # change threshold, if necessary
            maxi = _np.max(_np.abs(diff2))
            thres = maxi*1e-2 if maxi < thres else thres

            # identify zeros in 2 derivative and norm. configs indices
            ind = _np.where(_np.abs(diff2) < thres)[0]
            dind = _np.diff(ind)
            ind2 = _np.where(dind > 1)[0]
            idd_bef = _np.r_[ind[ind2], diff2.size-1]
            idd_aft = _np.r_[0, ind[ind2+1]]

        # calculate slopes, offsets and interceptions to interpolations
        slopes, offsets = list(), list()
        for beg, fin in zip(idd_aft, idd_bef):
            _x, _y = times[beg:fin], wfm[beg:fin]
            if not _x.size:
                continue
            off, slp = _np.polynomial.polynomial.polyfit(_x, _y, 1)
            slopes.append(slp)
            offsets.append(off)
        time_inter = -1 * _np.diff(offsets)/_np.diff(slopes)
        w_inter = slopes[:-1]*time_inter + offsets[:-1]
        if time_inter.size:
            time_inter2 = _np.round(time_inter, decimals=3)
            w_inter2 = _np.interp(time_inter2, time_inter, w_inter)
            time_inter, w_inter = time_inter2, w_inter2

        if self._use_rampconfig_times:
            if self._use_straigth_estim:
                w1_inter = slopes[1:]*nconf_times + offsets[1:]
                w2_inter = slopes[:-1]*nconf_times + offsets[:-1]
                w_inter = (w1_inter + w2_inter)/2
            else:
                w_inter = _np.interp(nconf_times, times, wfm)

            aux_t, aux_w = time_inter, w_inter
            val1 = _np.where(_np.logical_and(
                ~_np.isnan(aux_t), ~_np.isnan(aux_w)))[0]
            val2 = _np.where(_np.logical_and(
                _np.diff(-aux_t[val1]) <= 0,
                _np.diff(aux_t[val1]) >= 0))[0]
            time_inter = nconf_times[val1[val2]]
            w_inter = aux_w[val1[val2]]
            if aux_w.size and ~_np.isnan(aux_w[-1]):
                time_inter = _np.r_[time_inter, nconf_times[-1]]
                w_inter = _np.r_[w_inter, aux_w[-1]]
        return time_inter, w_inter

    def _get_initial_guess(self, threshold):
        # get time instants and normalized strengths where there are
        # normalized configurations
        problems = False
        ps2time2strg = dict()
        times = set()
        for psname, wfm in self._ps2wfms.items():
            time = wfm.times
            wfm = wfm.strengths

            time_inter, w_inter = self._calc_nconf_times(
                time, wfm, threshold)
            ps2time2strg[psname] = {
                i: w for i, w in zip(time_inter, w_inter)}
            times.update(time_inter)
            if not all(time_inter == sorted(time_inter)) or \
                    (time_inter.size == 1 and time_inter[0] == _np.nan):
                problems = True
                break
        if problems and threshold == BONormListFactory._THRESHOLD_DEFAULT:
            return False
        if problems:
            raise Exception(
                'Could not generate normalized configurations!')

        # sort and verify times
        times = sorted(times)
        if _np.any([time > self._duration for time in times]) or \
                _np.any([time < 0 for time in times]):
            raise Exception(
                'Could not generate normalized configurations!')
        if not times:
            times = {0.0, }

        # generate dict of normalized configs
        norm_configs_dict = dict()
        for time in times:
            energy = self._dipole.get_strength_from_time(time)
            nconf = dict()
            nconf['label'] = ' {:.4f}GeV'.format(energy)
            for dip in self._PSNAME_DIPOLES:
                nconf[dip] = energy
            for psname in self.psnames:
                if time in ps2time2strg[psname].keys():
                    nconf[psname] = ps2time2strg[psname][time]
                else:
                    wfm = self._ps2wfms[psname]
                    nconf[psname] = wfm.get_strength_from_time(time)
            norm_configs_dict['{:.3f}'.format(time)] = nconf
        return norm_configs_dict

    def _generate_nconf_dict(self):
        init_guess = self._get_initial_guess(
            BONormListFactory._THRESHOLD_DEFAULT)
        if not init_guess:
            init_guess = self._get_initial_guess(
                BONormListFactory._THRESHOLD_PROBLEM)

        self._norm_configs_dict = init_guess

        if self.precision_reached[0]:
            return

        self._order = list(init_guess.keys())

        if self._opt_global:
            nrows = len(self._psnames)+1 if self._opt_times else \
                len(self._psnames)
            ncols = len(self._order)

            init_params = _np.zeros((nrows, ncols))
            if self._opt_times:
                init_params[-1] = [float(t) for t in self._order]
            for row, psname in enumerate(self._psnames):
                for col, str_time in enumerate(self._order):
                    init_params[row][col] = init_guess[str_time][psname]

            final_params = least_squares(
                self._err_func_global, init_params.flatten(), method='lm').x
            final_params = _np.reshape(final_params, (nrows, ncols))

            final_times = final_params[-1] if self._opt_times else \
                [float(t) for t in self._order]

            nconfs = dict()
            for col, str_time_orig in enumerate(self._order):
                str_time = '{:.3f}'.format(final_times[col])
                nconfs[str_time] = dict()
                nconfs[str_time]['label'] = init_guess[str_time_orig]['label']
                for dip in self._PSNAME_DIPOLES:
                    energy = init_guess[str_time_orig][dip]
                    nconfs[str_time][dip] = energy
                for row, psname in enumerate(self._psnames):
                    nconfs[str_time][psname] = final_params[row][col]
        else:
            nconfs = dict()
            for psn in self._psnames:
                init_params = [init_guess[st][psn] for st in self._order]
                final_params = least_squares(
                    self._err_func_individual, init_params,
                    args=(psn,), method='lm').x
                for i, str_time in enumerate(self._order):
                    if str_time not in nconfs:
                        nconfs[str_time] = dict()
                        nconfs[str_time]['label'] = \
                            init_guess[str_time]['label']
                        for dip in self._PSNAME_DIPOLES:
                            energy = init_guess[str_time][dip]
                            nconfs[str_time][dip] = energy
                    nconfs[str_time][psn] = final_params[i]
        self._norm_configs_dict = nconfs

    def _err_func_individual(self, params, psname):
        nc_strengths = params
        nc_times = [float(t) for t in self._order]
        error = self._get_err_vector(psname, nc_times, nc_strengths)
        return error

    def _err_func_global(self, params):
        nrows = len(self._psnames)+1 if self._opt_times else \
            len(self._psnames)
        ncols = len(self._order)
        aux_params = _np.reshape(params, (nrows, ncols))
        if self._opt_times:
            aux_params = aux_params[:, _np.argsort(aux_params[-1, :])]
        nc_times = aux_params[-1] if self._opt_times else \
            [float(t) for t in self._order]

        error = list()
        for psn in self._psnames:
            nc_strengths = aux_params[self._psnames.index(psn)]
            ind_error = self._get_err_vector(psn, nc_times, nc_strengths)
            error.extend(list(ind_error))
        error = _np.array(error)
        return error

    def _get_err_vector(self, psname, nc_times, nc_strengths):
        if self._opt_metric == 'strength':
            error = self.calc_psname_strength_error(
                psname, nc_times, nc_strengths)
            loss_factor = self._get_loss_factor(psname)
            error /= loss_factor
        else:
            error = self.calc_psname_current_error(
                psname, nc_times, nc_strengths)
        return error

    def _get_loss_factor(self, psname):
        if psname.dev in ('QF', 'QD'):
            return BONormListFactory._LOSS_FACTOR_QUADS
        if psname.dev in ('SF', 'SD'):
            return BONormListFactory._LOSS_FACTOR_SEXTS
        return BONormListFactory._LOSS_FACTOR_CORRS


class BORFRampFactory:
    """Class to rebuild RF ramp parameters from machine state."""

    V_2_KV = 1e-3
    _DevName = 'BR-RF-DLLRF-01'
    _ppties = {
        'bottom_duration': _DevName+':RmpTs1-SP',
        'rampup_duration': _DevName+':RmpTs2-SP',
        'top_duration': _DevName+':RmpTs3-SP',
        'rampdown_duration': _DevName+':RmpTs4-SP',
        'bottom_voltage': 'RA-RaBO01:RF-LLRF:RmpAmpVCavBot-SP',
        'top_voltage': 'RA-RaBO01:RF-LLRF:RmpAmpVCavTop-SP',
        'bottom_phase': _DevName+':RmpPhsBot-SP',
        'top_phase': _DevName+':RmpPhsTop-SP',
    }
    _PVs = dict()

    def __init__(self):
        """."""
        self._rf_params = None
        self._create_pvs()

    @property
    def rf_params(self):
        """Return RF parameters rebuilt from PV values."""
        self._rf_params = self._generate_rf_params()
        return self._rf_params

    # ----- private methods -----

    def _create_pvs(self):
        pvs = dict()
        for param, pvname in BORFRampFactory._ppties.items():
            pvs[param] = _PV(pvname, connection_timeout=TIMEOUT_CONN)
        BORFRampFactory._PVs = pvs

    def _generate_rf_params(self):
        for pvobj in BORFRampFactory._PVs.values():
            pvobj.wait_for_connection()
            if not pvobj.connected:
                raise Exception('There are not connected PVs!')

        rf_params = dict()
        for param in BORFRampFactory._ppties:
            val = BORFRampFactory._PVs[param].value
            if 'voltage' in param:
                val = val*BORFRampFactory.V_2_KV
            rf_params[param] = val
        return rf_params


class BOTIRampFactory:
    """Class to rebuild TI ramp parameters from machine state."""

    _EVG = _LLTimeSearch.get_evg_name()
    _triggers = {
        'BO-Glob:TI-Mags-Fams', 'BO-Glob:TI-Mags-Corrs',
        'BO-Glob:TI-LLRF-Rmp',
        'LI-01:TI-EGun-SglBun', 'LI-01:TI-EGun-MultBun',
        'BO-48D:TI-EjeKckr', 'SI-01SA:TI-InjDpKckr',
    }
    _events = {
        _EVG+':Linac', _EVG+':InjBO', _EVG+':RmpBO', _EVG+':InjSI',
        _EVG+':Study',
    }
    _PVs = dict()

    def __init__(self):
        """."""
        self._ti_params = None
        BOTIRampFactory._create_pvs()

    @property
    def ti_params(self):
        """Return TI parameters rebuilt from PV values."""
        self._ti_params = self._generate_ti_params()
        return self._ti_params

    # ----- private methods -----

    @staticmethod
    def _create_pvs():
        pvs = dict()
        for trig in BOTIRampFactory._triggers:
            for ppty in {'Src-Sts', 'Delay-RB'}:
                pvname = trig + ':' + ppty
                pvs[pvname] = _PV(pvname, connection_timeout=TIMEOUT_CONN)

        for event in BOTIRampFactory._events:
            pvname = event + 'Delay-RB'
            pvs[pvname] = _PV(pvname, connection_timeout=TIMEOUT_CONN)

        egun_sb_sts_pvname = 'LI-01:EG-PulsePS:singleselstatus'
        pvs[egun_sb_sts_pvname] = _PV(
            egun_sb_sts_pvname, connection_timeout=TIMEOUT_CONN)

        BOTIRampFactory._PVs = pvs

    def _generate_ti_params(self):
        _pvs = BOTIRampFactory._PVs
        for pvobj in _pvs.values():
            pvobj.wait_for_connection()
            if not pvobj.connected:
                raise Exception('There are not connected PVs!')

        ti_params = dict()

        # injection time
        egun_src_idx = _pvs['LI-01:TI-EGun-SglBun:Src-Sts'].value
        egun_src_strs = _pvs['LI-01:TI-EGun-SglBun:Src-Sts'].enum_strs
        egun_src_val = egun_src_strs[egun_src_idx]
        egun_src_dly_pvn = \
            BOTIRampFactory._EVG+':'+egun_src_val+'Delay-RB'
        egun_src_dly = _pvs[egun_src_dly_pvn].value
        egun_dly = _pvs['LI-01:TI-EGun-SglBun:Delay-RB'].value \
            if _pvs['LI-01:EG-PulsePS:singleselstatus'].value \
            else _pvs['LI-01:TI-EGun-MultBun:Delay-RB'].value
        ti_params['injection_time'] = (egun_src_dly + egun_dly)/1e3

        # ejection time
        ejekckr_src_idx = _pvs['BO-48D:TI-EjeKckr:Src-Sts'].value
        ejekckr_src_strs = _pvs['BO-48D:TI-EjeKckr:Src-Sts'].enum_strs
        ejekckr_src_val = ejekckr_src_strs[ejekckr_src_idx]
        ejekckr_src_dly_pvn = \
            BOTIRampFactory._EVG+':'+ejekckr_src_val+'Delay-RB'
        ejekckr_src_dly = _pvs[ejekckr_src_dly_pvn].value
        ejekckr_dly = _pvs['BO-48D:TI-EjeKckr:Delay-RB'].value
        ti_params['ejection_time'] = (ejekckr_src_dly + ejekckr_dly)/1e3

        # mags delay
        ps_rmp_dly_fam = _pvs['BO-Glob:TI-Mags-Fams:Delay-RB'].value
        ti_params['ps_ramp_delay'] = (ps_rmp_dly_fam)/1e3

        # rf delay
        rf_rmp_dly = _pvs['BO-Glob:TI-LLRF-Rmp:Delay-RB'].value
        ti_params['rf_ramp_delay'] = (rf_rmp_dly)/1e3

        return ti_params



"""BO tune correction configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict)
    }
    return _dict


# Tune Correction Response Matrix for Booster
#
# | DeltaTuneX |   | m00  m01 |   | KL BO-Fam:PS-QF |
# |            | = |          | * |                 |
# | DeltaTuneY |   | m10  m11 |   | KL BO-Fam:PS-QD |
#
# Correction Matrix of Additional Method
# (obtained from lnls_calc_tunecorr_params routine)
#   m(0,0)   m(0,1)
#   m(1,0)   m(1,1)
#
# Nominal KLs
# [quadrupole_order  QF  QD]

_template_dict = {
    'matrix': [[0.0, 0.0],
               [0.0, 0.0]],
    'nominal KLs': [0.0, 0.0],
}
